# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from unittest import mock
import pytest

from mozphab import exceptions, mozphab
from mozphab.commands import reorganise


@pytest.mark.parametrize(
    "phids,transactions",
    [
        # No change
        (("A", "A"), {}),
        (("ABC", "ABC"), {}),
        (([], ["A"]), {}),
        # Abandon
        (
            ("ABC", "A"),
            {
                "A": [{"type": "children.remove", "value": ["B"]}],
                "B": [
                    {"type": "children.remove", "value": ["C"]},
                    {"type": "abandon", "value": True},
                ],
                "C": [{"type": "abandon", "value": True}],
            },
        ),
        (
            ("ABC", "B"),
            {
                "A": [
                    {"type": "children.remove", "value": ["B"]},
                    {"type": "abandon", "value": True},
                ],
                "B": [{"type": "children.remove", "value": ["C"]}],
                "C": [{"type": "abandon", "value": True}],
            },
        ),
        (
            ("ABC", "C"),
            {
                "A": [
                    {"type": "children.remove", "value": ["B"]},
                    {"type": "abandon", "value": True},
                ],
                "B": [
                    {"type": "children.remove", "value": ["C"]},
                    {"type": "abandon", "value": True},
                ],
            },
        ),
        # Reorder
        (
            ("AB", "BA"),
            {
                "A": [{"type": "children.remove", "value": ["B"]}],
                "B": [{"type": "children.set", "value": ["A"]}],
            },
        ),
        (
            ("ABC", "BC"),
            {
                "A": [
                    {"type": "children.remove", "value": ["B"]},
                    {"type": "abandon", "value": True},
                ]
            },
        ),
        (
            ("ABC", "ACB"),
            {
                "A": [{"type": "children.set", "value": ["C"]}],
                "B": [{"type": "children.remove", "value": ["C"]}],
                "C": [{"type": "children.set", "value": ["B"]}],
            },
        ),
        (
            ("ABC", "BAC"),
            {
                "A": [{"type": "children.set", "value": ["C"]}],
                "B": [{"type": "children.set", "value": ["A"]}],
            },
        ),
        (
            ("ABC", "CAB"),
            {
                "B": [{"type": "children.remove", "value": ["C"]}],
                "C": [{"type": "children.set", "value": ["A"]}],
            },
        ),
        (
            ("ABC", "CBA"),
            {
                "A": [{"type": "children.remove", "value": ["B"]}],
                "B": [{"type": "children.set", "value": ["A"]}],
                "C": [{"type": "children.set", "value": ["B"]}],
            },
        ),
        # Insert
        (("ABC", "DABC"), {"D": [{"type": "children.set", "value": ["A"]}]}),
        (
            ("ABC", "ADBC"),
            {
                "A": [{"type": "children.set", "value": ["D"]}],
                "D": [{"type": "children.set", "value": ["B"]}],
            },
        ),
        (
            ("ABC", "ABDC"),
            {
                "B": [{"type": "children.set", "value": ["D"]}],
                "D": [{"type": "children.set", "value": ["C"]}],
            },
        ),
        (
            ("ABC", "BCAD"),
            {
                "A": [{"type": "children.set", "value": ["D"]}],
                "C": [{"type": "children.set", "value": ["A"]}],
            },
        ),
        (([], ["A", "B"]), {"A": [{"type": "children.set", "value": ["B"]}]}),
        # Insert and reorder
        (
            ("ABC", "DCAB"),
            {
                "B": [{"type": "children.remove", "value": ["C"]}],
                "C": [{"type": "children.set", "value": ["A"]}],
                "D": [{"type": "children.set", "value": ["C"]}],
            },
        ),
        (
            ("ABC", "CDAB"),
            {
                "B": [{"type": "children.remove", "value": ["C"]}],
                "C": [{"type": "children.set", "value": ["D"]}],
                "D": [{"type": "children.set", "value": ["A"]}],
            },
        ),
        (
            ("ABC", "CADB"),
            {
                "A": [{"type": "children.set", "value": ["D"]}],
                "B": [{"type": "children.remove", "value": ["C"]}],
                "C": [{"type": "children.set", "value": ["A"]}],
                "D": [{"type": "children.set", "value": ["B"]}],
            },
        ),
        (
            ("ABC", "CABD"),
            {
                "B": [{"type": "children.set", "value": ["D"]}],
                "C": [{"type": "children.set", "value": ["A"]}],
            },
        ),
        # Nothing in common
        (("A", "B"), {"A": [{"type": "abandon", "value": True}]}),
    ],
)
def test_prepare_transactions(phids, transactions):
    assert reorganise.stack_transactions(*phids) == transactions


@pytest.mark.parametrize(
    "stacks,expected",
    [
        (({"A": None}, [{"rev-id": 1, "rev-phid": "A"}]), (["A"], ["A"])),
        (({"B": None}, [{"rev-id": 1, "rev-phid": "A"}]), (["B"], ["A"])),
        (
            ({"A": "B", "B": None}, [{"rev-id": 1, "rev-phid": "A"}]),
            (["A", "B"], ["A"]),
        ),
        (
            (
                {"A": None},
                [{"rev-id": 1, "rev-phid": "A"}, {"rev-id": 2, "rev-phid": "B"}],
            ),
            (["A"], ["A", "B"]),
        ),
    ],
)
@mock.patch("mozphab.commands.reorganise.stack_transactions")
@mock.patch("mozphab.conduit.ConduitAPI.check")
@mock.patch("mozphab.commands.reorganise.augment_commits_from_body")
@mock.patch("mozphab.conduit.ConduitAPI.get_stack")
@mock.patch("mozphab.conduit.ConduitAPI.get_revisions")
@mock.patch("mozphab.conduit.ConduitAPI.ids_to_phids")
@mock.patch("mozphab.conduit.ConduitAPI.phids_to_ids")
@mock.patch("mozphab.conduit.ConduitAPI.edit_revision")
def test_reorg_calling_stack_transactions(
    _edit_revision,
    _phid2id,
    m_id2phid,
    _get_revs,
    m_remote_stack,
    _augment_commits,
    _check,
    m_trans,
    git,
    stacks,
    expected,
):
    class Args:
        yes = True

    phabstack, commits = stacks
    m_remote_stack.return_value = phabstack
    mozphab.conduit.set_repo(git)
    mozphab.conduit.repo.commit_stack = mock.Mock()
    mozphab.conduit.repo.commit_stack.return_value = commits
    m_id2phid.return_value = [c["rev-phid"] for c in commits]
    reorganise.reorganise(git, Args())
    m_trans.assert_called_once_with(*expected)


@mock.patch("mozphab.conduit.ConduitAPI.check")
def test_conduit_broken(m_check):
    m_check.return_value = False
    with pytest.raises(exceptions.Error) as e:
        reorganise.reorganise(None, None)

    assert str(e.value) == "Failed to use Conduit API"


@mock.patch("mozphab.conduit.ConduitAPI.check")
@mock.patch("mozphab.commands.reorganise.augment_commits_from_body")
def test_commits_invalid(_augment, _check, git):
    mozphab.conduit.set_repo(git)
    mozphab.conduit.repo.commit_stack = mock.Mock()
    mozphab.conduit.repo.commit_stack.return_value = []
    with pytest.raises(exceptions.Error) as e:
        reorganise.reorganise(git, None)

    assert str(e.value) == "Failed to find any commits to reorganise."

    mozphab.conduit.repo.commit_stack.return_value = [{"rev-id": None, "name": "A"}]
    with pytest.raises(exceptions.Error) as e:
        reorganise.reorganise(git, None)

    error = str(e.value)
    assert error.startswith("Found new commit in the local stack: A.")


@mock.patch("mozphab.conduit.ConduitAPI.get_revisions")
@mock.patch("mozphab.conduit.ConduitAPI.check")
@mock.patch("mozphab.commands.reorganise.augment_commits_from_body")
@mock.patch("mozphab.conduit.ConduitAPI.get_stack")
@mock.patch("mozphab.conduit.ConduitAPI.phids_to_ids")
@mock.patch("mozphab.commands.reorganise.walk_llist")
@mock.patch("mozphab.commands.reorganise.logger")
def test_remote_stack_invalid(
    m_logger, m_walk, m_ids, m_stack, _augment, _check, _call, git
):
    mozphab.conduit.set_repo(git)
    mozphab.conduit.repo.commit_stack = mock.Mock()

    m_stack.return_value = {"A": "B"}
    m_ids.return_value = ["A", "B", "C"]
    mozphab.conduit.repo.commit_stack.return_value = [{"rev-id": 1, "name": "A"}]
    m_walk.side_effect = exceptions.Error("TEST")
    with pytest.raises(exceptions.Error) as e:
        reorganise.reorganise(git, None)

    assert str(e.value) == "TEST"
    m_logger.error.assert_called_once_with(
        "Remote stack is not linear.\nDetected stack:\nA <- B <- C"
    )


@pytest.mark.parametrize(
    "llist,error_text",
    (
        (dict(a="b", c=None), "Multiple heads"),
        (dict(a="b", b=None, c="d", d=None), "Multiple heads"),
        (dict(), "Failed to find head"),
        (dict(a="b", b="c", c="a"), "Failed to find head"),
        (dict(a="b", b="c", c="b"), "Dependency loop"),
    ),
)
def test_walk_llist_errors(llist, error_text):
    walk = reorganise.walk_llist
    with pytest.raises(exceptions.Error) as e:
        walk(llist)

    assert str(e.value).startswith(error_text)


@pytest.mark.parametrize(
    "llist,nodes",
    (
        (dict(a=None), ["a"]),
        (dict(a="b", b=None), ["a", "b"]),
        (dict(a="b"), ["a", "b"]),
        (dict(b="a", a=None), ["b", "a"]),
        (dict(a="b", b="c", c=None), ["a", "b", "c"]),
    ),
)
def test_walk_llist(llist, nodes):
    assert reorganise.walk_llist(llist) == nodes


@pytest.mark.parametrize(
    "params,result",
    (
        ([], {}),
        (["A"], {"A": None}),
        (["A", "B"], {"A": "B", "B": None}),
        (["A", "A"], {"A": None}),
    ),
)
def test_to_llist(params, result):
    assert reorganise.to_llist(params) == result



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import copy
import os
from unittest import mock
import time

from .conftest import hg_out, git_out

from mozphab import mozphab


@mock.patch("mozphab.conduit.ConduitAPI.get_revisions")
@mock.patch("mozphab.conduit.ConduitAPI.get_diffs")
@mock.patch("mozphab.conduit.ConduitAPI.call")
@mock.patch("mozphab.conduit.ConduitAPI.get_successor_phids")
@mock.patch("mozphab.conduit.ConduitAPI.get_ancestor_phids")
@mock.patch("builtins.print")
def test_patch_raw(
    m_print,
    m_ancestor_phids,
    m_get_successor_phids,
    m_call_conduit,
    m_get_diffs,
    m_get_revs,
    in_process,
    hg_repo_path,
    git_repo_path,
):
    def init(path):
        os.chdir(path)
        m_get_successor_phids.return_value = []
        m_get_revs.return_value = [REV_1]
        m_ancestor_phids.return_value = []
        m_get_diffs.return_value = {"PHID-DIFF-1": DIFF_1}
        m_get_revs.side_effect = None
        m_ancestor_phids.side_effect = None
        m_call_conduit.side_effect = None

    # `patch --raw` generates different output for hg vs git

    # git tests
    init(git_repo_path)

    m_print.reset_mock()
    m_call_conduit.return_value = PATCH_1_DIFF
    mozphab.main(["patch", "D1", "--raw"], is_development=True)
    assert m_print.call_args_list == [mock.call(PATCH_1_DIFF)]

    m_print.reset_mock()
    m_call_conduit.return_value = BIN_PATCH_DIFF
    mozphab.main(["patch", "D1", "--raw"], is_development=True)
    assert m_print.call_args_list == [mock.call(BIN_PATCH_DIFF)]

    m_print.reset_mock()
    m_call_conduit.return_value = BIN_PATCH_DIFF
    mozphab.main(["patch", "D1", "--raw"], is_development=True)
    assert m_print.call_args_list == [mock.call(BIN_PATCH_DIFF)]

    m_print.reset_mock()
    m_get_revs.side_effect = ([REV_2], [REV_1])
    m_ancestor_phids.side_effect = [["PHID-REV-1"], []]
    m_get_diffs.return_value = {"PHID-DIFF-1": DIFF_1, "PHID-DIFF-2": DIFF_2}
    m_call_conduit.side_effect = [PATCH_1_DIFF, PATCH_2_DIFF]
    mozphab.main(["patch", "D2", "--raw"], is_development=True)
    assert m_print.call_args_list == [mock.call(PATCH_1_DIFF), mock.call(PATCH_2_DIFF)]

    # hg tests
    init(hg_repo_path)

    m_print.reset_mock()
    m_call_conduit.return_value = PATCH_1_DIFF
    mozphab.main(["patch", "D1", "--raw"], is_development=True)
    assert m_print.call_args_list == [mock.call(PATCH_1_META + PATCH_1_DIFF)]

    m_print.reset_mock()
    m_call_conduit.return_value = BIN_PATCH_DIFF
    mozphab.main(["patch", "D1", "--raw"], is_development=True)
    assert m_print.call_args_list == [mock.call(BIN_PATCH_META + BIN_PATCH_DIFF)]

    m_print.reset_mock()
    m_call_conduit.return_value = BIN_PATCH_DIFF
    mozphab.main(["patch", "D1", "--raw"], is_development=True)
    assert m_print.call_args_list == [mock.call(BIN_PATCH_META + BIN_PATCH_DIFF)]

    m_print.reset_mock()
    m_get_revs.side_effect = ([REV_2], [REV_1])
    m_ancestor_phids.side_effect = [["PHID-REV-1"], []]
    m_get_diffs.return_value = {"PHID-DIFF-1": DIFF_1, "PHID-DIFF-2": DIFF_2}
    m_call_conduit.side_effect = [PATCH_1_DIFF, PATCH_2_DIFF]
    mozphab.main(["patch", "D2", "--raw"], is_development=True)
    assert m_print.call_args_list == [
        mock.call(PATCH_1_META + PATCH_1_DIFF),
        mock.call(PATCH_2_META + PATCH_2_DIFF),
    ]


@mock.patch("mozphab.conduit.ConduitAPI.get_revisions")
@mock.patch("mozphab.conduit.ConduitAPI.get_diffs")
@mock.patch("mozphab.conduit.ConduitAPI.call")
@mock.patch("mozphab.conduit.ConduitAPI.get_successor_phids")
@mock.patch("mozphab.conduit.ConduitAPI.get_ancestor_phids")
def test_patch_no_commit(
    m_ancestor_phids,
    m_get_successor_phids,
    m_call_conduit,
    m_get_diffs,
    m_get_revs,
    in_process,
    hg_repo_path,
):
    m_get_successor_phids.return_value = []
    m_get_revs.return_value = [REV_1]
    m_ancestor_phids.return_value = []
    m_get_diffs.return_value = {"PHID-DIFF-1": DIFF_1}
    m_call_conduit.side_effect = [
        dict(),
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="hg"))]),
        PATCH_1_DIFF,
    ]

    mozphab.main(["patch", "D1", "--no-commit"], is_development=True)
    assert [".arcconfig", ".hg", "X"] == sorted(os.listdir(str(hg_repo_path)))
    test_file = hg_repo_path / "X"
    assert "a\n" == test_file.read_text()
    result = hg_out("log", "-G")
    assert "@  changeset:   0:" in result
    test_file.unlink()

    m_call_conduit.side_effect = [BIN_PATCH_DIFF]
    mozphab.main(["patch", "D1", "--no-commit"], is_development=True)
    assert [".arcconfig", ".hg", "sample.bin"] == sorted(os.listdir(str(hg_repo_path)))
    test_file = hg_repo_path / "sample.bin"
    test_file.unlink()

    m_get_revs.side_effect = ([REV_2], [REV_1])
    m_ancestor_phids.side_effect = [["PHID-REV-1"], []]
    m_get_diffs.return_value = {"PHID-DIFF-1": DIFF_1, "PHID-DIFF-2": DIFF_2}
    m_call_conduit.side_effect = [PATCH_1_DIFF, PATCH_2_DIFF]
    mozphab.main(["patch", "D2", "--no-commit"], is_development=True)
    assert [".arcconfig", ".hg", "X"] == sorted(os.listdir(str(hg_repo_path)))
    test_file = hg_repo_path / "X"
    assert "b\n" == test_file.read_text()
    test_file.unlink()


@mock.patch("mozphab.conduit.ConduitAPI.get_revisions")
@mock.patch("mozphab.conduit.ConduitAPI.get_diffs")
@mock.patch("mozphab.conduit.ConduitAPI.call")
@mock.patch("mozphab.conduit.ConduitAPI.get_successor_phids")
@mock.patch("mozphab.conduit.ConduitAPI.get_ancestor_phids")
def test_git_patch_with_commit(
    m_ancestor_phids,
    m_get_successor_phids,
    m_call_conduit,
    m_get_diffs,
    m_get_revs,
    in_process,
    git_repo_path,
):
    m_get_successor_phids.return_value = []
    sha = git_out("rev-parse", "HEAD").rstrip("\n")
    diff_1 = copy.deepcopy(DIFF_1)
    diff_1["fields"]["refs"][0]["identifier"] = sha
    m_get_revs.return_value = [REV_1]
    m_ancestor_phids.return_value = []
    m_get_diffs.return_value = {"PHID-DIFF-1": diff_1}
    m_call_conduit.side_effect = [
        dict(),
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="git"))]),
        PATCH_1_DIFF,
    ]

    mozphab.main(["patch", "D1", "--apply-to", "here"], is_development=True)
    assert [".arcconfig", ".git", "X"] == sorted(os.listdir(str(git_repo_path)))
    test_file = git_repo_path / "X"
    assert "a\n" == test_file.read_text()
    result = git_out("log", "--all", "--format=[%at] %an <%ae>%n%s %P%n%b")
    assert 1 == result.count("[1547806078] user <author@example.com>")
    assert 1 == result.count("title R1")
    assert 1 == result.count("Differential Revision: http://example.test/D1")
    result = git_out("branch")
    assert "* phab-D1" in result

    time.sleep(1)  # to ensure the patch is applied with a different timestamp
    m_call_conduit.side_effect = [PATCH_1_DIFF]
    mozphab.main(["patch", "D1"], is_development=True)
    assert [".arcconfig", ".git", "X"] == sorted(os.listdir(str(git_repo_path)))
    test_file = git_repo_path / "X"
    assert "a\n" == test_file.read_text()
    result = git_out("log", "--all", "--format=[%at] %an <%ae>%n%s %P%n%b")
    assert 2 == result.count("[1547806078] user <author@example.com>")
    assert 2 == result.count("title R1")
    assert 2 == result.count("Differential Revision: http://example.test/D1")
    result = git_out("branch")
    assert "* phab-D1_1" in result
    assert "  phab-D1" in result

    time.sleep(1)
    m_get_revs.return_value = [REV_BIN]
    m_get_diffs.return_value = {"PHID-DIFF-3": diff_1}
    m_call_conduit.side_effect = [BIN_PATCH_DIFF]
    mozphab.main(["patch", "D3"], is_development=True)
    assert [".arcconfig", ".git", "sample.bin"] == sorted(
        os.listdir(str(git_repo_path))
    )
    result = git_out("log", "--all", "--format=[%at] %an <%ae>%n%s %P%n%b")
    assert 3 == result.count("[1547806078] user <author@example.com>")
    assert 1 == result.count("title BIN")
    assert 1 == result.count("Differential Revision: http://example.test/D3")
    assert 2 == result.count("title R1")
    assert 2 == result.count("Differential Revision: http://example.test/D1")
    result = git_out("branch")
    assert "* phab-D3" in result
    assert "  phab-D1_1" in result
    assert "  phab-D1" in result

    time.sleep(1)
    m_get_revs.side_effect = ([REV_2], [REV_1])
    m_ancestor_phids.side_effect = [["PHID-REV-1"], []]
    m_get_diffs.return_value = {"PHID-DIFF-1": diff_1, "PHID-DIFF-2": DIFF_2}
    m_call_conduit.side_effect = [PATCH_1_DIFF, PATCH_2_DIFF]
    mozphab.main(["patch", "D2"], is_development=True)
    assert [".arcconfig", ".git", "X"] == sorted(os.listdir(str(git_repo_path)))
    test_file = git_repo_path / "X"
    assert "b\n" == test_file.read_text()
    result = git_out("log", "--all", "--format=[%at] %an <%ae>%n%s %P%n%b")
    assert 5 == result.count("[1547806078] user <author@example.com>")
    assert 3 == result.count("title R1")
    assert 3 == result.count("Differential Revision: http://example.test/D1")
    assert 1 == result.count("Differential Revision: http://example.test/D2")
    assert 1 == result.count("title BIN")
    assert 1 == result.count("Differential Revision: http://example.test/D3")
    result = git_out("branch")
    assert "* phab-D2" in result
    assert "  phab-D3" in result
    assert "  phab-D1_1" in result
    assert "  phab-D1" in result

    time.sleep(1)
    m_get_revs.side_effect = ([REV_BIN],)
    m_get_diffs.return_value = {"PHID-DIFF-3": diff_1}
    m_call_conduit.side_effect = (PATCH_UTF8,)
    mozphab.main(["patch", "D4"], is_development=True)
    path = git_repo_path / "X"
    with path.open(encoding="utf-8") as f:
        line = f.readline().rstrip()

    assert line == "\u0105"


@mock.patch("mozphab.conduit.ConduitAPI.get_revisions")
@mock.patch("mozphab.conduit.ConduitAPI.get_diffs")
@mock.patch("mozphab.conduit.ConduitAPI.call")
@mock.patch("mozphab.conduit.ConduitAPI.get_successor_phids")
@mock.patch("mozphab.conduit.ConduitAPI.get_ancestor_phids")
def test_hg_patch_with_commit(
    m_ancestor_phids,
    m_get_successor_phids,
    m_call_conduit,
    m_get_diffs,
    m_get_revs,
    in_process,
    hg_repo_path,
):
    m_get_successor_phids.return_value = []
    m_get_revs.return_value = [REV_1]
    m_ancestor_phids.return_value = []
    m_get_diffs.return_value = {"PHID-DIFF-1": DIFF_1}
    m_call_conduit.side_effect = [
        dict(),
        dict(data=[dict(phid="PHID-REPO-1", fields=dict(vcs="hg"))]),
        PATCH_1_DIFF,
    ]

    mozphab.main(["patch", "D1", "--apply-to", "here"], is_development=True)
    assert [".arcconfig", ".hg", "X"] == sorted(os.listdir(str(hg_repo_path)))
    test_file = hg_repo_path / "X"
    assert "a\n" == test_file.read_text()
    result = hg_out("log", "-G")
    assert "@  changeset:   1:" in result
    assert "|  bookmark:    phab-D1" in result
    assert "|  user:        user <author@example.com>" in result
    assert "|  date:        Fri Jan 18" in result
    assert "|  summary:     title R1" in result
    assert "o  changeset:   0:" in result

    m_call_conduit.side_effect = [BIN_PATCH_DIFF, 67]
    mozphab.main(["patch", "D1"], is_development=True)
    assert [".arcconfig", ".hg", "sample.bin"] == sorted(os.listdir(str(hg_repo_path)))
    result = hg_out("log", "-G")
    assert "@  changeset:   2" in result
    assert "|  bookmark:    phab-D1_1" in result
    assert "|  parent:      0" in result
    assert "| o  changeset:   1" in result
    assert "|/   bookmark:    phab-D1" in result

    testfile = hg_repo_path / "unknown"
    testfile.write_text("not added to repository")
    m_get_revs.side_effect = ([REV_2], [REV_1])
    m_ancestor_phids.side_effect = [["PHID-REV-1"], []]
    m_get_diffs.return_value = {"PHID-DIFF-1": DIFF_1, "PHID-DIFF-2": DIFF_2}
    m_call_conduit.side_effect = [PATCH_1_DIFF, PATCH_2_DIFF]
    mozphab.main(["patch", "D2"], is_development=True)
    assert [".arcconfig", ".hg", "X", "unknown"] == sorted(
        os.listdir(str(hg_repo_path))
    )
    test_file = hg_repo_path / "X"
    assert "b\n" == test_file.read_text()
    result = hg_out("log", "-G")
    assert "@  changeset:   3:" in result
    assert "|  bookmark:    phab-D2" in result
    assert "|  parent:      1" in result
    assert "|  summary:     title R2" in result
    assert "| o  changeset:   2" in result
    assert "| |  bookmark:    phab-D1_1" in result
    assert "| |  parent:      0" in result
    assert "o |  changeset:   1" in result
    assert "|/   bookmark:    phab-D1" in result

    # Same, but this time request a specific bookmark name.
    m_get_revs.side_effect = ([REV_2], [REV_1])
    m_ancestor_phids.side_effect = [["PHID-REV-1"], []]
    m_get_diffs.return_value = {"PHID-DIFF-1": DIFF_1, "PHID-DIFF-2": DIFF_2}
    m_call_conduit.side_effect = [PATCH_1_DIFF, PATCH_2_DIFF]
    mozphab.main(["patch", "--name=myfeature", "D2"], is_development=True)
    result = hg_out("log", "-G")
    assert "@  changeset:   3:" in result
    assert "|  bookmark:    myfeature" in result
    assert "|  parent:      1" in result
    assert "|  summary:     title R2" in result
    assert "| o  changeset:   2" in result
    assert "| |  bookmark:    phab-D1_1" in result
    assert "| |  parent:      0" in result
    assert "o |  changeset:   1" in result
    assert "|/   bookmark:    phab-D1" in result


REV_1 = dict(
    id=1,
    phid="PHID-REV-1",
    fields=dict(title="title R1", summary="\u0105", diffPHID="PHID-DIFF-1"),
)

REV_2 = dict(
    id=2,
    phid="PHID-REV-2",
    fields=dict(title="title R2", summary="\u0105", diffPHID="PHID-DIFF-2"),
)

REV_BIN = dict(
    id=3,
    phid="PHID-REV-3",
    fields=dict(title="title BIN", summary="\u0105", diffPHID="PHID-DIFF-3"),
)
ATTACHMENTS = dict(
    commits=dict(
        commits=[dict(author=dict(name="user", email="author@example.com", epoch=None))]
    )
)

DIFF_1 = dict(
    fields=dict(dateCreated=1547806078, refs=[dict(identifier="0", type="base")]),
    id=1,
    attachments=ATTACHMENTS,
)

DIFF_2 = dict(fields=dict(dateCreated=1547806078), id=2, attachments=ATTACHMENTS)

PATCH_1_META = """\
# HG changeset patch
# User user <author@example.com>
# Date 1547806078 0
title R1

ą

Differential Revision: http://example.test/D1
"""
PATCH_1_DIFF = """\
diff --git a/X b/X
new file mode 100644
--- /dev/null
+++ b/X
@@ -0,0 +1 @@
+a

"""

PATCH_2_META = """\
# HG changeset patch
# User user <author@example.com>
# Date 1547806078 0
title R2

ą

Differential Revision: http://example.test/D2

Depends on D1
"""
PATCH_2_DIFF = """\
diff --git a/X b/X
--- a/X
+++ b/X
@@ -1 +1 @@
-a
+b

"""

PATCH_UTF8 = """\
diff --git a/X b/X
new file mode 100644
--- /dev/null
+++ b/X
@@ -0,0 +1 @@
+\u0105

"""

BIN_PATCH_META = """\
# HG changeset patch
# User user <author@example.com>
# Date 1547806078 0
title R1

ą

Differential Revision: http://example.test/D1
"""
BIN_PATCH_DIFF = """\
diff --git a/sample.bin b/sample.bin
new file mode 100644
index 0000000000000000000000000000000000000000..ea9bc1d9ecfd0d99001e388c9b70dd588b1ba265
GIT binary patch
literal 3072
zc$@(M4FB_Yh{OHEZg>2xm*xN6(kD$&7ARO{=H`W*Ipz@K$nP4h#B8`Cds^=fQaXDv
z?=(-#Z2R&`E%_xW_I=L<+NMQTx!Gy{f83c%vLn9L7LM$PONedE?i!GByxKK;NhBTu
z?%izKL+>y0EhgJ2-(bu=FY*Z)I(sonWXCkbfhS-@j(qc4R|~GWelS`6I4Qu4{I%cD
z&K5QiPO7xGd~*MfN0|JhuKrrn^!GI#ZKdnu#<Wsfv7NfcfN7YwU=PV;?rYzM(OA18
z4=nB)8V8+6%s}C%>ceR?4C8_D;Llr&A<*yZ1x=-IenIZXODh!M>?hoZ^HVD2oUxK>
zy$Z$Kkto+&pvVoT><u6j5p&ZX^82wZEEBO!qJRE7m2=$g+fmSZ^S2NG`#wM?__Zkc
zh^jFQHEFW)gcnsyDV-Vg_6dhL0+_w89pZnNMRXz@C*+C3sF4j5`5IXt-!}S;oLT6P
z!=`~a0h><~UM9xCUv&GLs<wbE<m)tUxxrG9<u~Oi19@{8r)7rz(GvA;X|2&7ChtUz
z43lnGfK$1Q-=qCpXbN1B(%Vy>X50f>;8C<!5s<^T9|gOl#*HF*x*LG2-!1_lfTY@K
zOeadX20y=SBGq##QnX$9<U6#~nH0E{zQTrckY2oO!^zA;xJ@GSlpl*5=M&x;pGy8u
zdY-iuF4~otMbm)Yv0w?BluPdFSeiWP3kSAllUq&XIH0^6`Kab@kbP*7vkD54w`LpG
zvOD!8=GL$L=p*?;Eq$%_&k99`2{e1nDzvX3>&HN)U1gxN8!ENhdkziA3C>s@>YE?|
zZt_l|dRVPXotsimG(hb}O7r_{Z~J5|H;}<Vv~>84lU*1Qz=BI)eT)N+av?V;OF$Vn
zzR%>YZ6hNOu=~>SzDA=Iq*xq?M~2Z)hZNYY<tC1hLUWjxd9^GLe$U|7LI9v9slQ1s
z5!D1t+v$nOA;@GouoQ*F99dP>%@hvPO1kWY7<$T<xlOldC)tZi1=OzRz=3~Es7r|7
z(a}h%4Ga2a@yj&9c&p|h<vmBuR%hk#1IG;|Ip#Txyw51ymXu!5V5KMr>et*o8Vwg~
zW&Z*W|1Ci%+A-Xm?{I>3E06@$FyGLdBw!9NeD7osq9~z&KAu^F{HDXroz7M!^`2my
zONI}5gk?joWF6px4O3@-8Z~;XA+sQ4*@{}5FmxF7P-#36Y^Rx-K_hI#+_tf~hGTeZ
z3p+z=x|hPE-S_SU+*9_L?@0v$n_VYzO5uvXHAhM#y_Xd}Wgh7f-Y!Z}Gp8fd?@K|)
z2+vHWbS&mdd=$mE0bfx2yh^(U&W@2Sw+q=<J|CE%$PR43thp2*uh&>|Odo2FrGZmv
zCR_#}9uI$r>0YU5TyIpS#8}g5d2StWWdKeU*qpK(^J0?miN)@0Utm((s1d)vA^c^F
z@c!Y_2_Nx+%&u@A#3vs2)N=E}wvfv4(_S+=Sq<N2g6tFSRmlOL&y5=6br?rPO{mG=
zzB0dQ5wBR6GgnGmENfV*3~#WOtK0jJFw~H$yR~<H-)-XzdfCLRCbu3%A@3o3tu?=!
zxAY(fe??#;yFg@!Ik(T>N;kL)PhYm879%l;LwaaMZv}5m#lu<tzc;kh-3*^fcQOHR
z@a#j^0CXzW5|qD_pM&8}z8yX*IRlML@867})Uv_LERkHWq1M&Mbn=e2Lau#7$((^g
z799D;*6a*c$+lV{Qy5&!#eh#NkMPH;??COPWnJ}b-3sH=ED_(AHEN!PD+Zau-&x5D
z(Wd#5@Zfv@MQYbzv$Qf@AUXdb!o}r@pjyf9013dCRfU@Z*7TA+JPGiaW4m0w6b8P4
zM&G<j9w|W+h;b7e!>Lp~RJiHrQ<MUTX+|yVd%CLk?Is%pzSPYq>t3+21Fa5oL=H!i
z58n_cDvD}qmwj9EDbmV)+rpc14WN84kR7pDNS4tDu0FC*X`aFBg7_tbtpykZ(HN{~
z5M-UaOiL#3M(k{r$opU$-|~jvdBS8Wam88Hv;fxB=TMCiI~6d8fx2gA2hf!L&5n`&
zAZ|SJMWHY~6F}5)#g1HiZ-_QouZ?l>2HtASEx>mf@e-z@?l4(kPO!hM^-E=KOTY8&
zd%jI#CL{-RO;qtT?G={4K=U^Xpy~ZR0>UyMpXjg-V1w~l7Q?I4d1PEEnIzzSva=0k
zORZQtZCPs7#Kh>{y@y`vt{Gsc3MGaC=h)1vGC-MiCdJ_sM>0y?N_1!{{yY{DBF9=J
z0!#D7uUv}|3_SPHB-kQ@_}qgkSZ{xvjw?Ysjh(mGydwgjh;lOq<kxl)0;L+j6yyQ^
zxqpB3N?W{4?wB-No@3-5TV7XnPsAr@3RT==99&h4nI^h(8pZL7x8FDO<i7l<ZQ<^`
z$#Do(^RV=u%Z*G0$Z`3`6j13PvJstRs+&3i^}R1rXskeey%fiasV2`v>VnMfgbEm%
zUIrST$EOlnoUz-x968DXsa1ZM24JywIOf6MwWzT#(8nOQ{kk4`n?0-1(wTT5sHzL;
z;~IiyrL<)#aJR`D%c4#&*fXHVCpKXsMpCVV-TB~LRKF6Z=69<Grn_<|v8x{!Z>uu*
zgdk;_xXuuDP;cXKAWX$h?Ko}d7upmL_zvxRp-r>4Id)zXqX^YtXD^EjHJ^F0`dNWP
zs+4TeBJqR#dm45-E$J^sz-&HW;$nd@N2WS053g&zC4E?hZL<+r=E139Q_R}JMdB5$
zt>19MKimW1#SfoCv3DM22E@(Xi2Vnxkvh^6dz!}aYx3rz?{#!XWp$`#>GgZ_5wS&K
zT7q=YtKeZBijFv|kOQ&Lc?1pwp8WjQLR7I08|w%hvT?9SwK?msO{IuB?7%upqm_)?
z+yxN?OAz=L?GT<tv9=!pB{I9eL0sKJkhar^CkIGD<I6d}>x<B_;9j%Nk^~iXFXf_y
z-i_<GZK$zv%M3SWjq^*0x$(8vzy4rV%ai&|Dt^$#&FL^TbBR+KVMr2~ZKEbYjjzQ3
zNt&vpT=>b;VsRL+YT%SQK1)Fb2`oR4s4<b^sC*x^T|)>`*xVg73_hx<V;A)7;Z!wd
z3wh{uMEe=t5&~Rt`bYO^c%_v{8{@!!8|EWNb6)+<v5G90i+OD2S;kb(L`!mU*_&+l
z1k5?xHi~^zJp)Xf)oOdH30}9mKU4`uiN^dOX0!mg$7r4*g4`J@l}&T;I8I>|d(<2#
z584<=U<%u-+mq&S-iO95huSWUx|9Y>+_G&4I}Nh-UR-NA(NF&~7Oh%gLjt9Qh<I_M
z7}qh0K3rXqbFc?ndSErj9uv~)HZ7$L#=fO2#;fb(yo+Bc)OJpm%|COo;%^IP#YYnj
z(#ZUDYepE1rP(+r{~}`6QuJuzY3fTo=cmwz*T804BU>WNEi1)R4N(%pNum5^8!M;|
z<|ETY)5V3dh{8fWe!GFqIiO8_#L@5!%25QL1*xn3VP3;pOmg66O$`?#R_H>sFCxNj
z_Th9u;hA48I;beSyoTs&C1Au$P}V?6`JNa|fcO;a7yyzz*9hHvg{p#@#Ojw}{ZCu$
zv-sW>b>ADSQ|4L#az}|*O&Mw>sCza9TU?~@3K<j7w`)IDUhd+~Pe-3ONWgmA{z6fX
zs4A6v6@A)p{d?5}Qz%@2p@?l*p-f&-)}@7^NtK6W36>GTs4%(PjPje@jD~oVF%x=_
z_soAyV9rO884hHA$f#}`1rhO|=|KlLwHptE+e9^K;kg80Xi+PpGPWNx^A<B^$>zZB
zJT3<>SH*C@7Rpq>mSn$;{7wAbBiG#su`GOVW2q+6IrYgaP#$%qNRF5Ko~l{bKG6bb
zN8NP0lSg8%Y9I8X@EsM!_xEU|oNkEgQ2Sm$AWF=N9NX&B?qoym_Tn)aag67VDBoO!
z>GA5XA`C6J{yd+F73#Gc?<sD^+BjimUrM)eB=;`tV{s)a^@8X|d$lk%4*qq(z}It6
znUE~MzjO42m7DH*4A7Zvj(<0+Nx*$eQtR}S^2)PA)NM`lWBlg>F>%~m)gG1$!Kc6t
O!n%`!v-e$`d$Yhe6Zzr*

literal 0
Hc$@<O00001


"""



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from pathlib import Path
import subprocess

from .conftest import find_script_path


ROOT = Path(__file__).resolve().parent.parent
PY_FILES = sorted(
    str(f)
    for f in list(ROOT.glob("*.py"))
    + list((ROOT / "mozphab").glob("**/*.py"))
    + list((ROOT / "tests").glob("**/*.py"))
)


def test_black():
    subprocess.check_call([find_script_path("black"), "--check"] + PY_FILES)


def test_flake8():
    subprocess.check_call(
        [find_script_path("flake8")]
        + ["--max-line-length=88"]
        + ["--ignore=E203,W503"]
        + ["--disable-noqa"]
        + PY_FILES
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import errno
import logging
import socket
import ssl
import urllib.error

import hglib.error
import sentry_sdk
from pkg_resources import get_distribution
from sentry_sdk.integrations.logging import LoggingIntegration

from .exceptions import CommandError


def init_sentry():
    distribution = get_distribution("MozPhab")

    sentry_logging = LoggingIntegration(
        level=logging.INFO,
        event_level=None,
        # The default event_level is logging.ERROR, which will report any
        # "logging.error(...)" call to Sentry.  However, we respond to
        # incorrect usage with "logging.error(...)" messages, which we don't
        # want to report to Sentry.
    )
    sentry_sdk.init(
        dsn=(
            "https://f2dcfa028ddb4540b5d64a855d480909@o1069899.ingest.sentry.io/6250015"
        ),
        integrations=[sentry_logging],
        release=distribution.version,
    )


def report_to_sentry(e):
    if (
        # As we don't capture stdout/stderr, failures when running hg/git are not
        # actionable.  Some cases of CommandError never need reporting (eg. failure
        # to apply a patch).  Mercurial's command server can throw these as ServerErrors
        isinstance(e, CommandError)
        or isinstance(e, hglib.error.ServerError)
        # SSLCertVerification errors are caused by a misconfigured Python install.
        or isinstance(e, ssl.SSLCertVerificationError)
        # Network unreachable
        or (isinstance(e, OSError) and e.errno == errno.ENETUNREACH)
        # Network timeout
        or isinstance(e, TimeoutError)
        # DNS resolution failures
        or isinstance(e, socket.gaierror)
        # urllib throws URLError on SSL verification, network unreachable, etc
        or isinstance(e, urllib.error.URLError)
        # Connection resets are transient
        or isinstance(e, ConnectionResetError)
        # Ctrl-C can manifest in a number of ways
        or isinstance(e, KeyboardInterrupt)
        or isinstance(e, BrokenPipeError)
    ):
        return

    sentry_sdk.capture_exception(e)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import copy
from unittest import mock
import unittest
import uuid

from callee import Contains

from mozphab.commands import submit
from mozphab import environment, exceptions, helpers, mozphab, repository

from .conftest import search_diff, search_rev


def reviewers_dict(reviewers=None):
    return dict(
        request=reviewers[0] if reviewers else [],
        granted=reviewers[1] if reviewers else [],
    )


def commit(
    bug_id=None, reviewers=None, body="", name="", title="", rev_id=None, wip=False
):
    return {
        "name": name,
        "title": title,
        "bug-id": bug_id,
        "reviewers": reviewers_dict(reviewers),
        "has-reviewers": bool(reviewers),
        "body": body,
        "rev-id": rev_id,
        "node": uuid.uuid4().hex,
        "submit": True,
        "wip": wip,
    }


# noinspection PyPep8Naming,PyBroadException
class Commits(unittest.TestCase):
    def _assertNoError(self, callableObj, *args, **kwargs):
        try:
            callableObj(*args, **kwargs)
        except exceptions.Error as e:
            self.fail("%s raised" % e)

    def _assertError(self, callableObj, expected, *args, **kwargs):
        try:
            callableObj(*args, **kwargs)
        except exceptions.Error as e:
            if expected != str(e).strip():
                self.fail("%s not raised" % expected)
            return
        except Exception as e:
            self.fail("%s raised" % e)
        self.fail("%s failed to raise Error" % callableObj)

    @mock.patch("mozphab.conduit.ConduitAPI.check_for_invalid_reviewers")
    @mock.patch("mozphab.conduit.ConduitAPI.get_revisions")
    @mock.patch("mozphab.conduit.ConduitAPI.whoami")
    def test_commit_validation(self, m_whoami, m_get_revs, check_reviewers):
        check_reviewers.return_value = []
        repo = repository.Repository("", "", "dummy")
        check = repo.check_commits_for_submit

        self._assertNoError(check, [])
        self._assertNoError(check, [commit("1", (["r"], []))])
        self._assertNoError(
            check,
            [
                commit("1", (["r1"], [])),
                commit("2", (["r1"], [])),
                commit("3", (["r1", "r2"], [])),
            ],
        )
        self._assertNoError(check, [commit("1", None)])
        self._assertNoError(check, [commit("1", (["r"], [])), commit("1", None)])

        self._assertError(check, "- missing bug-id", [commit(None, (["r"], []))])
        self._assertNoError(check, [commit(None, (["r"], []))], require_bug=False)
        self._assertError(check, "- missing bug-id", [commit("", (["r"], []))])
        self._assertError(
            check,
            "- missing bug-id",
            [commit("1", (["r"], [])), commit("", (["r"], []))],
        )
        self._assertNoError(
            check,
            [commit("1", (["r"], [])), commit("", (["r"], []))],
            require_bug=False,
        )

        self._assertError(
            check,
            "- contains arc fields",
            [commit("1", (["r"], []), body="Summary: blah\nReviewers: r")],
        )

        m_whoami.return_value = dict(phid="PHID-1")
        m_get_revs.return_value = [dict(fields=dict(authorPHID="PHID-1"))]
        self._assertNoError(check, [commit(bug_id=1, rev_id=1)])
        m_whoami.return_value = dict(phid="PHID-2")
        self._assertNoError(check, [commit(bug_id=1, rev_id=1)])

    @mock.patch("mozphab.conduit.ConduitAPI.check_for_invalid_reviewers")
    def test_invalid_reviewers_fails_the_stack_validation_check(self, check_reviewers):
        class Args:
            def __init__(self):
                self.force = False

        def fail_gonzo(reviewers):
            # Replace the check_for_invalid_reviewers() function with something that
            # fails if "gonzo" is in the reviewers list.
            if "gonzo" in reviewers["request"]:
                return [dict(name="gonzo")]
            elif "goober" in reviewers["request"]:
                return [dict(name="goober", until="string")]
            else:
                return []

        check_reviewers.side_effect = fail_gonzo
        repo = repository.Repository("", "", "dummy")
        repo.args = Args()

        self._assertError(
            repo.check_commits_for_submit,
            "- gonzo is not a valid reviewer's name",
            (
                # Build a stack with an invalid reviewer in the middle.
                [
                    commit("1", (["alice"], [])),
                    commit("2", (["bob", "gonzo"], [])),
                    commit("3", (["charlie"], [])),
                ]
            ),
        )

        self._assertError(
            repo.check_commits_for_submit,
            "- goober is not available until string",
            (
                # Build a stack with an unavailable reviewer in the middle.
                [
                    commit("1", (["alice"], [])),
                    commit("2", (["bob", "goober"], [])),
                    commit("3", (["charlie"], [])),
                ]
            ),
        )

        repo.args.force = True
        self._assertNoError(
            repo.check_commits_for_submit,
            (
                [
                    commit("1", (["alice"], [])),
                    commit("2", (["bob", "goober"], [])),
                    commit("3", (["charlie"], [])),
                ]
            ),
        )

    @mock.patch("mozphab.mozphab.conduit.get_revisions")
    @mock.patch("mozphab.conduit.ConduitAPI.check_for_invalid_reviewers")
    def test_validate_duplicate_revision(self, check_reviewers, get_revisions):
        check_reviewers.return_value = []
        get_revisions.return_value = [True]

        repo = repository.Repository("", "", "dummy")

        self._assertNoError(
            repo.check_commits_for_submit,
            (
                [
                    commit("1", (["r"], []), name="a"),
                    commit("2", (["r"], []), name="b"),
                    commit("3", (["r"], []), name="c"),
                ]
            ),
        )

        self._assertNoError(
            repo.check_commits_for_submit,
            (
                [
                    commit("1", (["r"], []), rev_id="1", name="a"),
                    commit("2", (["r"], []), rev_id="2", name="b"),
                    commit("3", (["r"], []), rev_id="3", name="c"),
                ]
            ),
        )

        self._assertError(
            repo.check_commits_for_submit,
            "Phabricator revisions should be unique, "
            "but the following commits refer to the same one (D1):\n"
            "* a\n"
            "* c",
            (
                [
                    commit("1", (["r"], []), rev_id="1", name="a"),
                    commit("2", (["r"], []), rev_id="2", name="b"),
                    commit("3", (["r"], []), rev_id="1", name="c"),
                ]
            ),
        )

        self._assertError(
            repo.check_commits_for_submit,
            "Phabricator revisions should be unique, "
            "but the following commits refer to the same one (D1):\n"
            "* a\n"
            "* c"
            "\n\n\n"
            "Phabricator revisions should be unique, "
            "but the following commits refer to the same one (D2):\n"
            "* b\n"
            "* d",
            (
                [
                    commit("1", (["r"], []), rev_id="1", name="a"),
                    commit("2", (["r"], []), rev_id="2", name="b"),
                    commit("3", (["r"], []), rev_id="1", name="c"),
                    commit("4", (["r"], []), rev_id="2", name="d"),
                ]
            ),
        )

    def test_commit_preview(self):
        build = helpers.build_commit_title

        self.assertEqual(
            "Bug 1, blah, r=turnip",
            build(commit("1", ([], ["turnip"]), title="bug 1, blah, r=turnip")),
        )
        self.assertEqual(
            "blah (Bug 1) r=turnip",
            build(commit("1", ([], ["turnip"]), title="blah (bug 1) r=turnip")),
        )
        self.assertEqual(
            "Bug 1 - blah r?turnip",
            build(commit("1", (["turnip"], []), title="blah r?turnip")),
        )

        self.assertEqual(
            "blah r=turnip", build(commit("", ([], ["turnip"]), title="blah r=turnip"))
        )
        self.assertEqual(
            "Bug 1 - blah", build(commit("1", None, title="Bug 1 - blah r?turnip"))
        )
        self.assertEqual(
            "Bug 1 - blah", build(commit("1", None, title="Bug 1 - blah r=turnip"))
        )
        self.assertEqual(
            "Bug 1 - helper_bug2.html",
            build(commit("1", None, title="Bug 1 - helper_bug2.html")),
        )

    @mock.patch("mozphab.helpers.build_commit_title")
    def test_update_commit_title_previews(self, m_build_commit_title):
        m_build_commit_title.side_effect = lambda x: x["title"] + " preview"
        commits = [dict(title="a"), dict(title="b")]
        helpers.update_commit_title_previews(commits)
        self.assertEqual(
            [
                {"title": "a", "title-preview": "a preview"},
                {"title": "b", "title-preview": "b preview"},
            ],
            commits,
        )

    def test_replace_request_reviewers(self):
        replace = helpers.replace_reviewers
        self.assertEqual("", replace("", reviewers_dict()))
        self.assertEqual("Title", replace("Title", reviewers_dict()))
        self.assertEqual(
            "Title\n\nr?one r=two", replace("Title\n\nr?one r=two", reviewers_dict())
        )
        self.assertEqual("r?one", replace("", reviewers_dict([["one"], []])))
        self.assertEqual("r?one", replace("r?one", reviewers_dict([["one"], []])))
        self.assertEqual("r?one,two", replace("", reviewers_dict([["one", "two"], []])))
        self.assertEqual(
            "Some Title r?one,two",
            replace("Some Title", reviewers_dict([["one", "two"], []])),
        )
        self.assertEqual(
            "Title r?one\n\nDescr\niption",
            replace("Title\n\nDescr\niption", reviewers_dict([["one"], []])),
        )
        self.assertEqual(
            "Title r?one,two\n\nr?two",
            replace("Title\n\nr?two", reviewers_dict([["one", "two"], []])),
        )

        self.assertEqual("Title", replace("Title r?one", reviewers_dict()))
        self.assertEqual("Title", replace("Title r?one,two", reviewers_dict()))
        self.assertEqual("Title", replace("Title r?one r?two", reviewers_dict()))
        self.assertEqual(
            "Title r?one!", replace("Title r?one!", reviewers_dict([["one!"], []]))
        )
        self.assertEqual(
            "Title r?one", replace("Title r?one!", reviewers_dict([["one"], []]))
        )
        self.assertEqual(
            "Title r?one!", replace("Title r?one", reviewers_dict([["one!"], []]))
        )
        self.assertEqual(
            "Title r?one", replace("Title r?one", reviewers_dict([["one"], []]))
        )
        self.assertEqual(
            "Title r?one one", replace("Title r? one", reviewers_dict([["one"], []]))
        )
        self.assertEqual(
            "Title r?one,two",
            replace("Title r?one,two", reviewers_dict([["one", "two"], []])),
        )
        self.assertEqual(
            "Title r?one,two",
            replace("Title r?two", reviewers_dict([["one", "two"], []])),
        )
        self.assertEqual(
            "Title r?one,two",
            replace("Title r?one r?two", reviewers_dict([["one", "two"], []])),
        )
        self.assertEqual(
            "Title r?one", replace("Title r=one", reviewers_dict([["one"], []]))
        )
        self.assertEqual(
            "Title r?one", replace("Title r=one,two", reviewers_dict([["one"], []]))
        )
        self.assertEqual(
            "Title r?one,two",
            replace("Title r=one,two", reviewers_dict([["one", "two"], []])),
        )
        self.assertEqual(
            "Title r?one,two",
            replace("Title r=one r=two", reviewers_dict([["one", "two"], []])),
        )
        self.assertEqual(
            "Title r?one,two",
            replace("Title r?one r=two", reviewers_dict([["one", "two"], []])),
        )

    def test_replace_granted_reviewers(self):
        replace = helpers.replace_reviewers
        self.assertEqual("r=one", replace("", reviewers_dict([[], ["one"]])))
        self.assertEqual("r=one", replace("r=one", reviewers_dict([[], ["one"]])))
        self.assertEqual("r=one,two", replace("", reviewers_dict([[], ["one", "two"]])))
        self.assertEqual(
            "Some Title r=one,two",
            replace("Some Title", reviewers_dict([[], ["one", "two"]])),
        )
        self.assertEqual(
            "Title r=one\n\nDescr\niption",
            replace("Title\n\nDescr\niption", reviewers_dict([[], ["one"]])),
        )
        self.assertEqual(
            "Title r=one,two\n\nr?two",
            replace("Title\n\nr?two", reviewers_dict([[], ["one", "two"]])),
        )
        self.assertEqual("Title", replace("Title r=one", reviewers_dict()))
        self.assertEqual("Title", replace("Title r=one,two", reviewers_dict()))
        self.assertEqual("Title", replace("Title r=one r=two", reviewers_dict()))
        self.assertEqual(
            "Title r=one", replace("Title r=one", reviewers_dict([[], ["one"]]))
        )
        self.assertEqual(
            "Title r=one!", replace("Title r=one!", reviewers_dict([[], ["one!"]]))
        )
        self.assertEqual(
            "Title r=one", replace("Title r=one!", reviewers_dict([[], ["one"]]))
        )
        self.assertEqual(
            "Title r=one!", replace("Title r=one", reviewers_dict([[], ["one!"]]))
        )
        self.assertEqual(
            "Title r=one one", replace("Title r= one", reviewers_dict([[], ["one"]]))
        )
        self.assertEqual(
            "Title r=one,two",
            replace("Title r=one,two", reviewers_dict([[], ["one", "two"]])),
        )
        self.assertEqual(
            "Title r=one,two",
            replace("Title r=two", reviewers_dict([[], ["one", "two"]])),
        )
        self.assertEqual(
            "Title r=one,two",
            replace("Title r=one r=two", reviewers_dict([[], ["one", "two"]])),
        )
        self.assertEqual(
            "Title r=one", replace("Title r?one", reviewers_dict([[], ["one"]]))
        )
        self.assertEqual(
            "Title r=one", replace("Title r?one,two", reviewers_dict([[], ["one"]]))
        )
        self.assertEqual(
            "Title r=one,two",
            replace("Title r?one,two", reviewers_dict([[], ["one", "two"]])),
        )
        self.assertEqual(
            "Title r=one,two",
            replace("Title r?one r?two", reviewers_dict([[], ["one", "two"]])),
        )
        self.assertEqual(
            "Title r=one,two",
            replace("Title r=one r?two", reviewers_dict([[], ["one", "two"]])),
        )

    def test_replace_mixed_reviewers(self):
        replace = helpers.replace_reviewers
        self.assertEqual(
            "Title r?one r=two", replace("Title", reviewers_dict([["one"], ["two"]]))
        )
        self.assertEqual(
            "Title r?one r=two",
            replace("Title r=one r?two", reviewers_dict([["one"], ["two"]])),
        )
        self.assertEqual(
            "Title r?one r=two",
            replace("Title r?two r=one", reviewers_dict([["one"], ["two"]])),
        )
        self.assertEqual(
            "Title r?one,two r=three",
            replace("Title r=one r?two", reviewers_dict([["one", "two"], ["three"]])),
        )
        self.assertEqual(
            "Title r?one r=two,three",
            replace("Title r=one r?two", reviewers_dict([["one"], ["two", "three"]])),
        )

    @unittest.skip("These tests should pass we should fix the function")
    def test_badly_replaced_reviewers(self):
        replace = mozphab.replace_reviewers
        # r?two
        self.assertEqual("r?one", replace("r?two", reviewers_dict([["one"], []])))
        # r=one
        self.assertEqual("r?one", replace("r=one", reviewers_dict([["one"], []])))
        # r? one
        self.assertEqual("r?one one", replace("r? one", reviewers_dict([["one"], []])))
        # r?one,
        self.assertEqual("r?one", replace("r?one,", reviewers_dict([["one"], []])))
        # r?one
        self.assertEqual(
            "r?one,two", replace("r?one", reviewers_dict([["one", "two"], []]))
        )
        # r?one,two
        self.assertEqual("r?one", replace("r?one,two", reviewers_dict([["one"], []])))
        # Title r?one,two,,two
        self.assertEqual(
            "Title r?one,two",
            replace("Title r?one,,two", reviewers_dict([["one"], ["two"], []])),
        )
        # r?one
        self.assertEqual("", replace("r?one", reviewers_dict()))
        # r?one,two
        self.assertEqual("", replace("r?one,two", reviewers_dict()))
        # r?one
        self.assertEqual("", replace("r?one r?two", reviewers_dict()))
        # r?two
        self.assertEqual(
            "r?one,two", replace("r?two", reviewers_dict([["one", "two"], []]))
        )
        # r?one r?one,two
        self.assertEqual(
            "r?one,two", replace("r?one r?two", reviewers_dict([["one", "two"], []]))
        )
        # r=one
        self.assertEqual("r?one", replace("r=one", reviewers_dict([["one"], []])))
        # r=one,two
        self.assertEqual("r?one", replace("r=one,two", reviewers_dict([["one"], []])))
        # r=one,two
        self.assertEqual(
            "r?one,two", replace("r=one,two", reviewers_dict([["one", "two"], []]))
        )
        # r=one, r?one,two
        self.assertEqual(
            "r?one,two", replace("r=one r=two", reviewers_dict([["one", "two"], []]))
        )
        # r? one, r?one,two
        self.assertEqual(
            "r?one,two", replace("r?one r=two", reviewers_dict([["one", "two"], []]))
        )

        # Granted
        # r=two
        self.assertEqual("r=one", replace("r=two", reviewers_dict([[], ["one"]])))
        # r?one
        self.assertEqual("r=one", replace("r?one", reviewers_dict([[], ["one"]])))
        # r?one
        self.assertEqual(
            "r=one,two", replace("r=one", reviewers_dict([[], ["one", "two"]]))
        )
        # r?one,two
        self.assertEqual("r?one", replace("r?one,two", reviewers_dict([["one"], []])))

    @mock.patch("mozphab.commands.submit.conduit.get_revisions")
    @mock.patch("mozphab.commands.submit.conduit.get_diffs")
    @mock.patch("mozphab.commands.submit.conduit.whoami")
    @mock.patch("mozphab.commands.submit.logger")
    def test_show_commit_stack(
        self, mock_logger, m_whoami, m_get_diffs, m_get_revisions
    ):
        class Repository:
            phab_url = "http://phab"
            path = "x"
            api_url = "x"
            dot_path = "x"

        def _commit(
            name="aaa000",
            node="aaa000aaa000",
            title="A",
            rev=None,
            bug="1",
            bug_orig="1",
            granted=None,
            request=None,
            wip=False,
        ):
            granted = granted or []
            request = request or []
            return {
                "name": name,
                "node": node,
                "submit": True,
                "title-preview": title,
                "rev-id": rev,
                "bug-id-orig": bug_orig,
                "bug-id": bug,
                "reviewers": {"granted": granted, "request": request},
                "has-reviewers": len(granted + request) > 0,
                "wip": wip,
            }

        repo = Repository()
        submit.conduit.set_repo(repo)

        m_whoami.return_value = dict(phid="PHID-USER-1")
        m_get_revisions.return_value = [search_rev()]
        m_get_diffs.return_value = {"PHID-DIFF-1": search_diff()}

        submit.show_commit_stack([])
        self.assertFalse(mock_logger.info.called, "logger.info() shouldn't be called")
        self.assertFalse(
            mock_logger.warning.called, "logger.warning() shouldn't be called"
        )

        submit.show_commit_stack(
            [{"name": "aaa000", "title-preview": "A"}], validate=False
        )
        mock_logger.info.assert_called_with("%s %s %s", "(New)", "aaa000", "A")
        self.assertFalse(
            mock_logger.warning.called, "logger.warning() shouldn't be called"
        )
        mock_logger.reset_mock()

        submit.show_commit_stack([_commit(rev="12")], validate=False)
        mock_logger.info.assert_called_with("%s %s %s", "(D12)", "aaa000", "A")
        self.assertFalse(
            mock_logger.warning.called, "logger.warning() shouldn't be called"
        )
        mock_logger.reset_mock()

        submit.show_commit_stack(
            [_commit(), _commit(name="bbb000", title="B")],
            validate=False,
        )
        self.assertEqual(2, mock_logger.info.call_count)
        self.assertEqual(
            [
                mock.call("%s %s %s", "(New)", "bbb000", "B"),
                mock.call("%s %s %s", "(New)", "aaa000", "A"),
            ],
            mock_logger.info.call_args_list,
        )
        mock_logger.reset_mock()

        submit.show_commit_stack(
            [_commit(bug_orig="2", bug="1", granted=["alice"])], validate=True
        )
        mock_logger.info.assert_called_with("%s %s %s", "(New)", "aaa000", "A")
        mock_logger.warning.assert_called_with(
            "!! Bug ID changed from %s to %s", "2", "1"
        )
        mock_logger.reset_mock()
        submit.show_commit_stack([_commit()], validate=True)
        mock_logger.warning.assert_called_with(Contains("Missing reviewers"))
        mock_logger.reset_mock()

        submit.show_commit_stack(
            [_commit(rev="123")],
            validate=False,
            show_rev_urls=True,
        )
        mock_logger.warning.assert_called_with("-> %s/D%s", "http://phab", "123")

        # Do not update not changed commits
        m_get_revisions.reset_mock()
        m_get_revisions.return_value = [
            search_rev(),
            search_rev(rev="2", phid="PHID-REV-2", diff="PHID-DIFF-2"),
        ]
        # we're changing bug id in the first revision to 2
        submit.show_commit_stack(
            [
                _commit(rev="1", bug="2", bug_orig="1", granted=["alice"]),
                _commit(
                    name="bbb000",
                    node="bbb000bbb000",
                    title="B",
                    rev="2",
                    granted=["alice"],
                ),
            ],
            validate=True,
        )
        assert mock_logger.warning.call_args_list[1] == mock.call(
            "!! Bug ID in Phabricator revision will change from %s to %s", "1", "2"
        )
        assert m_get_revisions.call_count == 3
        mock_logger.reset_mock()

        m_whoami.return_value = dict(phid="PHID-USER-2")
        submit.show_commit_stack(
            [_commit(rev="1", granted=["alice"])],
            validate=True,
        )
        mock_logger.warning.assert_called_once_with(Contains("Commandeer"))

        # Information about not updating the commit if not changed
        mock_logger.reset_mock()
        m_whoami.return_value = dict(phid="PHID-USER-1")
        m_get_revisions.return_value = [search_rev(reviewers=["PHID-USER-2"])]
        m_get_diffs.return_value = {"PHID-DIFF-1": search_diff()}

        submit.show_commit_stack([_commit(rev="1", granted=["alice"])], validate=True)
        assert mock_logger.info.call_args_list[1] == mock.call(
            Contains("revision has not changed")
        )

        # Removing the WIP state from the revision without changing the commit's SHA1
        mock_logger.reset_mock()
        m_get_revisions.return_value = [search_rev(status="changes-planned")]
        submit.show_commit_stack([_commit(rev="1", granted=["alice"])], validate=True)
        mock_logger.warning.assert_called_once_with(
            Contains('"Changes Planned" status will change')
        )

        # Adding the WIP state to the revision without changing the commit's SHA1
        mock_logger.reset_mock()
        m_get_revisions.return_value = [search_rev()]
        submit.show_commit_stack(
            [_commit(rev="1", granted=["alice"], wip=True)], validate=True
        )
        mock_logger.warning.assert_called_with(
            Contains('status will change to "Changes Planned"')
        )

    @mock.patch("mozphab.commands.submit.update_commit_title_previews")
    def test_update_commits_from_args(self, m_update_title):
        def lwr(revs):
            return [r.lower() for r in revs]

        m_update_title.side_effect = lambda x: x
        update = submit.update_commits_from_args

        class Args:
            def __init__(
                self,
                reviewer=None,
                blocker=None,
                bug=None,
                wip=False,
                no_wip=False,
                command="submit",
            ):
                self.reviewer = reviewer
                self.blocker = blocker
                self.bug = bug
                self.wip = wip
                self.no_wip = no_wip
                self.command = command

        _commits = [
            {"title": "A", "reviewers": dict(granted=[], request=[]), "bug-id": None},
            {
                "title": "B",
                "reviewers": dict(granted=[], request=["one"]),
                "bug-id": "1",
            },
        ]

        # No change if noreviewer  args provided
        commits = copy.deepcopy(_commits)
        commits[1]["reviewers"]["granted"].append("two")
        with mock.patch("mozphab.commands.submit.config") as m_config:
            m_config.always_blocking = False
            update(commits, Args())
            self.assertEqual(
                commits,
                [
                    {
                        "title": "A",
                        "reviewers": dict(granted=[], request=[]),
                        "bug-id": None,
                        "has-reviewers": False,
                        "wip": True,
                    },
                    {
                        "title": "B",
                        "reviewers": dict(granted=["two"], request=["one"]),
                        "bug-id": "1",
                        "has-reviewers": True,
                        "wip": False,
                    },
                ],
            )

        # Adding and removing reviewers, forcing the bug ID
        commits = copy.deepcopy(_commits)
        update(commits, Args(reviewer=["two", "three"], bug="2"))
        assert commits[0]["title"] == "A"
        assert commits[0]["bug-id"] == "2"
        assert "two" in commits[0]["reviewers"]["granted"]
        assert "three" in commits[0]["reviewers"]["granted"]

        assert commits[1]["title"] == "B"
        assert commits[1]["bug-id"] == "2"
        assert "two" in commits[1]["reviewers"]["granted"]
        assert "three" in commits[1]["reviewers"]["granted"]

        # Removing duplicates
        commits = copy.deepcopy(_commits)
        update(
            commits,
            Args(
                reviewer=["Two", "two", "two!", "three", "Three", "THREE!"],
                blocker=["Two", "THREE!", "three", "two", "three"],
            ),
        )
        assert commits[0]["title"] == "A"
        assert commits[0]["bug-id"] is None
        assert "two!" in lwr(commits[0]["reviewers"]["granted"])
        assert "three!" in lwr(commits[0]["reviewers"]["granted"])

        assert commits[1]["title"] == "B"
        assert commits[1]["bug-id"] == "1"
        assert "two!" in lwr(commits[1]["reviewers"]["granted"])
        assert "three!" in lwr(commits[1]["reviewers"]["granted"])

        # Adding blocking reviewers via args
        commits = copy.deepcopy(_commits)
        commits[1]["reviewers"]["request"].append("three")
        commits[1]["reviewers"]["granted"].append("four")
        commits[1]["reviewers"]["granted"].append("five")
        update(
            commits,
            Args(
                reviewer=["one", "two!", "four"],
                blocker=["three", "four!"],
            ),
        )
        assert commits[0]["title"] == "A"
        assert commits[0]["bug-id"] is None
        assert "one" in lwr(commits[0]["reviewers"]["granted"])
        assert "two!" in lwr(commits[0]["reviewers"]["granted"])
        assert "three!" in lwr(commits[0]["reviewers"]["granted"])
        assert "four!" in lwr(commits[0]["reviewers"]["granted"])

        assert commits[1]["title"] == "B"
        assert commits[1]["bug-id"] == "1"
        assert "four!" in commits[1]["reviewers"]["granted"]
        assert "two!" in commits[1]["reviewers"]["granted"]
        assert "one" in commits[1]["reviewers"]["request"]
        assert "three!" in commits[1]["reviewers"]["request"]

        # reviewerless should result in WIP commits
        commits = copy.deepcopy(_commits)
        update(commits, Args())
        assert commits[0]["wip"]
        assert not commits[1]["wip"]

        # Force WIP
        commits = copy.deepcopy(_commits)
        update(commits, Args(wip=True))
        assert commits[0]["wip"]
        assert commits[1]["wip"]

        # reviewerless with --no-wip shouldn't be WIP
        commits = copy.deepcopy(_commits)
        update(commits, Args(no_wip=True))
        assert not commits[0]["wip"]
        assert not commits[1]["wip"]

        # Forcing blocking reviewers
        commits = copy.deepcopy(_commits)
        commits[1]["reviewers"]["granted"].append("two")
        with mock.patch("mozphab.commands.submit.config") as m_config:
            m_config.always_blocking = True
            update(commits, Args())
            self.assertEqual(
                commits,
                [
                    {
                        "title": "A",
                        "reviewers": dict(granted=[], request=[]),
                        "bug-id": None,
                        "has-reviewers": False,
                        "wip": True,
                    },
                    {
                        "title": "B",
                        "reviewers": dict(granted=["two!"], request=["one!"]),
                        "bug-id": "1",
                        "has-reviewers": True,
                        "wip": False,
                    },
                ],
            )

    def test_single_fails_with_end_rev(self):
        # --single is working with one SHA1 provided only
        repo = repository.Repository("", "", "dummy")

        class Args:
            def __init__(self, end_rev=environment.DEFAULT_END_REV):
                self.single = True
                self.end_rev = end_rev

        self._assertNoError(repo.set_args, Args())
        self._assertError(
            repo.set_args,
            "Option --single can be used with only one identifier.",
            Args("endrev"),
        )


class TestUpdateCommitSummary(unittest.TestCase):
    def test_update_revision_description(self):
        c = commit(
            rev_id="D123",
            title="hi!",
            body="hello!  µ-benchmarks are a thing.\n\n"
            "Differential Revision: http://phabricator.test/D123",
        )
        r = dict(fields=dict(title="", summary=""))
        t = []

        expected = [
            dict(type="title", value="hi!"),
            dict(type="summary", value="hello!  µ-benchmarks are a thing."),
        ]
        submit.update_revision_description(t, c, r)

        self.assertListEqual(t, expected)

    def test_update_revision_description_no_op(self):
        c = commit(
            rev_id="D123",
            title="hi!",
            body="hello!\n\nDifferential Revision: http://phabricator.test/D123",
        )
        r = dict(fields=dict(title="hi!", summary="hello!\n\n"))
        t = []

        expected = []
        submit.update_revision_description(t, c, r)

        self.assertListEqual(t, expected)

    @mock.patch("mozphab.conduit.ConduitAPI.get_users")
    def test_update_revision_reviewers(self, m_get_users):
        # From https://phabricator.services.mozilla.com/api/differential.revision.edit
        #
        # Example call format we are aiming for:
        #
        # $ echo '{
        #   "transactions": [
        #     {
        #       "type": "reviewers.set",
        #       "value": ["PHID-USER-1", "blocking(PHID-USER-2)"]
        #     }
        #   ],
        #   "objectIdentifier": "D8095"
        # }' | arc call-conduit --conduit-uri \
        #       https://phabricator.services.mozilla.com/ \
        #       --conduit-token <conduit-token> differential.revision.edit
        m_get_users.side_effect = (
            [
                dict(phid="PHID-USER-1"),
                dict(phid="PHID-USER-2"),
                dict(phid="PHID-USER-3"),
            ],
            [dict(phid="PHID-USER-1")],
            [dict(phid="PHID-USER-2"), dict(phid="PHID-USER-3")],
        )
        c = commit(rev_id="123", reviewers=[["alice", "bob!"], ["frankie!"]])
        t = []

        expected = [
            dict(
                type="reviewers.set",
                value=["PHID-USER-1", "blocking(PHID-USER-2)", "blocking(PHID-USER-3)"],
            )
        ]
        mozphab.conduit.update_revision_reviewers(t, c)

        self.assertEqual(
            m_get_users.call_args_list,
            [
                mock.call(["alice", "bob!", "frankie!"]),
                mock.call(["alice"]),
                mock.call(["bob", "frankie"]),
            ],
        )
        self.assertListEqual(t, expected)

    def test_parse_api_response_with_no_problems(self):
        # Response comes from running:
        # $ echo '{... (some valid update summary JSON) ...}' | \
        #   arc call-conduit differential.revision.edit
        api_response = (
            '{"error":null,"errorMessage":null,"response":{"object":{'
            '"id":56,"phid":"PHID-DREV-ke6jhbdnwd5chtnk2q5w"},'
            '"transactions":[{"phid":"PHID-XACT-DREV-itlxgx7rsjrcnta"}]}} '
        )
        self.assertEqual(None, helpers.parse_api_error(api_response))

    def test_parse_api_response_with_errors(self):
        # Error response from running:
        # $ echo '{}' | arc call-conduit differential.revision.edit
        api_response = (
            '{"error":"ERR-CONDUIT-CORE", '
            '"errorMessage":"ERR-CONDUIT-CORE: Parameter '
            '\\"transactions\\" is not a list of transactions.",'
            '"response":null} '
        )
        self.assertEqual(
            'ERR-CONDUIT-CORE: Parameter "transactions" is not a list of transactions.',
            helpers.parse_api_error(api_response),
        )

    def test_update_revision_no_bug_id(self):
        # Phabricator stores patches with no bug as having an empty string as the
        # bug ID. We should not explicitly update the bug-id if our bug id is "None".
        transactions = []
        submit.update_revision_bug_id(
            transactions, {"bug-id": None}, {"fields": {"bugzilla.bug-id": ""}}
        )
        self.assertEqual([], transactions)


if __name__ == "__main__":
    unittest.main()



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import subprocess

from shlex import quote

from .exceptions import CommandError
from .logger import logger


def debug_log_command(command):
    logger.debug("$ %s", " ".join(quote(s.replace("\n", r"\n")) for s in command))


def check_call(command, **kwargs):
    # wrapper around subprocess.check_call with debug output
    debug_log_command(command)
    kwargs["encoding"] = "UTF-8"
    try:
        subprocess.check_call(command, **kwargs)
    except subprocess.CalledProcessError as e:
        raise CommandError(
            "command '%s' failed to complete successfully" % command[0], e.returncode
        )


def check_call_by_line(command, cwd=None, never_log=False):
    # similar to check_call, yields for line-by-line processing
    debug_log_command(command)

    # Connecting the STDIN to the PIPE will make arc throw an exception on reading
    # user input
    process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stdin=subprocess.PIPE,
        cwd=cwd,
        universal_newlines=True,
    )
    try:
        for line in iter(process.stdout.readline, ""):
            line = line.rstrip()
            if not never_log:
                logger.debug("> %s", line)
            yield line
    finally:
        process.stdout.close()
        process.wait()

    if process.returncode:
        raise CommandError(
            "command '%s' failed to complete successfully" % command[0],
            process.returncode,
        )


def check_output(
    command,
    cwd=None,
    split=True,
    keep_ends=False,
    strip=True,
    never_log=False,
    stdin=None,
    stderr=None,
    env=None,
    search_error=None,
    expect_binary=False,
):
    # wrapper around subprocess.check_output with debug output and splitting
    debug_log_command(command)
    kwargs = dict(cwd=cwd, stdin=stdin, stderr=stderr)
    if not expect_binary:
        kwargs["universal_newlines"] = True
        kwargs["encoding"] = "UTF-8"

    if env:
        kwargs["env"] = env

    try:
        output = subprocess.check_output(command, **kwargs)
    except subprocess.CalledProcessError as e:
        if search_error:
            for err in search_error:
                if err["matching"] in e.output:
                    logger.error(err["message"])

        if e.output and not never_log:
            logger.debug(e.output)

        if e.stderr and not never_log:
            logger.debug(e.stderr)

        raise CommandError(
            "command '%s' failed to complete successfully" % command[0], e.returncode
        )

    if expect_binary:
        logger.debug("%s bytes of data received", len(output))
        return output

    if strip:
        output = output.rstrip()
    if output and not never_log:
        logger.debug(output)
    return output.splitlines(keep_ends) if split else output



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import base64
import datetime
import hashlib
import json
import os
import urllib.parse as url_parse
import urllib.request as url_request

from typing import (
    Any,
    Dict,
    List,
    Optional,
)

from mozphab import environment

from .environment import USER_AGENT
from .exceptions import (
    CommandError,
    Error,
    NonLinearException,
    NotFoundError,
)
from .helpers import (
    get_arcrc_path,
    read_json_field,
    revision_title_from_commit,
    strip_differential_revision,
)
from .logger import logger
from .simplecache import cache


CHECK_IN_NEEDED = "check-in_needed"

REVISION_STATUS_TO_TRANSACTION = {
    "abandoned": "abandon",
    "accepted": "accept",
    "changes-planned": "plan-changes",
    "draft": "draft",
    "needs-review": "request-review",
    "needs-revision": "reject",
}


def normalise_reviewer(reviewer: str, strip_group: bool = True) -> str:
    """This provide a canonical form of the reviewer for comparison."""
    reviewer = reviewer.rstrip("!").lower()
    if strip_group:
        reviewer = reviewer.lstrip("#")
    return reviewer


class ConduitAPIError(Error):
    """Raised when the Phabricator Conduit API returns an error response."""

    def __init__(self, msg: Optional[str] = None):
        super().__init__(f"Phabricator Error: {msg if msg else 'Unknown Error'}")


class ConduitAPI:
    def __init__(self):
        self.repo = None

    def set_repo(self, repo):
        self.repo = repo

    @property
    def repo_phid(self) -> str:
        return self.repo.phid

    def load_api_token(self) -> str:
        """Return an API Token for the given repository.

        Returns:
            API Token string
        """

        if "api_token" in cache:
            return str(cache.get("api_token"))

        token = read_json_field(
            [get_arcrc_path()], ["hosts", self.repo.api_url, "token"]
        )
        if not token:
            raise ConduitAPIError(environment.INSTALL_CERT_MSG)
        cache.set("api_token", token)
        return token

    def save_api_token(self, token: str):
        filename = get_arcrc_path()
        created = False
        try:
            with open(filename, "r", encoding="utf-8") as f:
                rc = json.load(f)
        except FileNotFoundError:
            rc = {}
            created = True

        rc.setdefault("hosts", {})
        rc["hosts"].setdefault(self.repo.api_url, {})
        rc["hosts"][self.repo.api_url]["token"] = token

        with open(filename, "w", encoding="utf-8") as f:
            json.dump(rc, f, sort_keys=True, indent=2)

        if created:
            os.chmod(filename, 0o600)

    def call(
        self, api_method: str, api_call_args: dict, *, api_token: Optional[str] = None
    ) -> dict:
        """Call Conduit API and return the JSON API call result.

        Args:
            api_method: The API method name to call, like 'differential.revision.edit'.
            api_call_args: JSON dict of call args to send.
            api_token: Use specific token to authentication instead of saved value.

        Returns:
            JSON API call result object

        Raises:
            ConduitAPIError if the API threw an error back at us.
        """
        req_args = self._build_request(
            method=api_method,
            args=api_call_args,
            token=api_token,
        )
        logger.debug("%s %s", req_args["url"], api_call_args)

        with url_request.urlopen(url_request.Request(**req_args)) as r:
            res = json.load(r)
        if res["error_code"]:
            raise ConduitAPIError(res.get("error_info", "Error %s" % res["error_code"]))
        return res["result"]

    def _build_request(
        self, *, method: str, args: dict, token: Optional[str]
    ) -> Dict[str, Any]:
        """Return dict with Request args for calling the specified conduit method."""
        return dict(
            url=url_parse.urljoin(self.repo.api_url, method),
            method="POST",
            headers={"User-Agent": USER_AGENT},
            data=url_parse.urlencode(
                {
                    "params": json.dumps(
                        {
                            **args,
                            "__conduit__": {"token": token or self.load_api_token()},
                        },
                        separators=(",", ":"),
                    ),
                    "output": "json",
                    "__conduit__": True,
                }
            ).encode(),
        )

    def ping(self) -> bool:
        """Sends a ping to the Phabricator server using `conduit.ping` API.

        Returns: `True` if no error, otherwise - `False`
        """
        try:
            self.call("conduit.ping", {})
        except ConduitAPIError as err:
            logger.error(err)
            return False
        except CommandError as err:
            logger.error(err)
            return False
        return True

    def check(self) -> bool:
        """Check if raw Conduit API can be used."""
        # Check if the cache file exists
        path = os.path.join(self.repo.dot_path, ".moz-phab_conduit-configured")
        if os.path.isfile(path):
            return True

        if self.ping():
            # Create the cache file
            with open(path, "a"):
                os.utime(path, None)
            return True

        return False

    def get_projects(self, slugs: List[str]) -> List[dict]:
        """Search for tags by hashtags."""
        response = self.call("project.search", dict(constraints=dict(slugs=slugs)))
        return response.get("data")

    def get_project_phid(self, slug: str) -> Optional[str]:
        projects = self.get_projects([slug])
        if not projects:
            return None

        return projects[0]["phid"]

    def ids_to_phids(self, rev_ids: List[int]) -> List[str]:
        """Convert revision ids to PHIDs.

        Parameters:
            rev_ids (list): A list of revision ids

        Returns:
            A list of PHIDs.
        """
        return [r["phid"] for r in self.get_revisions(ids=rev_ids)]

    def id_to_phid(self, rev_id: int) -> str:
        """Convert revision id to PHID."""
        phids = self.ids_to_phids([rev_id])
        if phids:
            return phids[0]

        raise NotFoundError("revision {} not found".format(rev_id))

    def phids_to_ids(self, phids: List[str]) -> List[str]:
        """Convert revision PHIDs to ids.

        Parameteres:
            phids (list): A list of PHIDs

        Returns:
            A list of ids.
        """
        return ["D{}".format(r["id"]) for r in self.get_revisions(phids=phids)]

    def phid_to_id(self, phid: str) -> str:
        """Convert revision PHID to id."""
        ids = self.phids_to_ids([phid])
        if ids:
            return ids[0]

        raise NotFoundError("revision {} not found".format(phid))

    def get_revisions(
        self, ids: Optional[List[int]] = None, phids: Optional[List[int]] = None
    ) -> List[dict]:
        """Get revisions info from Phabricator.

        Args:
            ids - list of revision ids
            phids - list of revision phids

        Returns a list of revisions ordered by ids or phids
        """
        if (ids and phids) or (ids is None and phids is None):
            raise ValueError("Internal Error: Invalid args to get_revisions")

        # Initialise depending on if we're passed revision IDs or PHIDs.
        if ids:
            ids = [str(rev_id) for rev_id in ids]
            phids_by_id = dict(
                [
                    (rev_id, cache.get("rev-id-%s" % rev_id))
                    for rev_id in ids
                    if "rev-id-%s" % rev_id in cache
                ]
            )
            found_phids = list(phids_by_id.values())
            query_field = "ids"
            query_values = [
                int(rev_id) for rev_id in set(ids) - set(phids_by_id.keys())
            ]

        else:
            phids_by_id = {}
            found_phids = phids.copy()
            query_field = "phids"
            query_values = set([phid for phid in phids if "rev-%s" % phid not in cache])

        # Revisions metadata keyed by PHID.
        revisions = dict(
            [
                (phid, cache.get("rev-%s" % phid))
                for phid in found_phids
                if "rev-%s" % phid in cache
            ]
        )

        # Query Phabricator if we don't have cached values for revisions.
        if query_values:
            api_call_args = {
                "constraints": {query_field: sorted(query_values)},
                "attachments": {"reviewers": True},
            }
            response = self.call("differential.revision.search", api_call_args)
            rev_list = response.get("data")

            for r in rev_list:
                phids_by_id[str(r["id"])] = r["phid"]
                revisions[r["phid"]] = r
                cache.set("rev-id-%s" % r["id"], r["phid"])
                cache.set("rev-%s" % r["phid"], r)

        # Return revisions in the same order requested.
        if ids:
            # Skip revisions for which we do not have a query result.
            return [
                revisions[phids_by_id[rev_id]]
                for rev_id in ids
                if rev_id in phids_by_id
            ]
        else:
            return [revisions[phid] for phid in phids]

    def get_diffs(
        self, ids: Optional[List[int]] = None, phids: Optional[List[str]] = None
    ) -> Dict[str, Dict]:
        """Get diffs from Phabricator.

        Args:
            ids - a list of diff IDs to pull
            phids - a list of diff PHIDs to pull

        Returns a dict of diffs identified by their PHID
        """
        if (ids and phids) or (ids is None and phids is None):
            raise ValueError("Internal Error: Invalid args to get_diffs")

        if ids:
            constraints = {"ids": list(set(ids))}
        else:
            constraints = {"phids": list(set(phids))}

        api_call_args = {
            "constraints": constraints,
            "attachments": {"commits": True},
        }
        response = self.call("differential.diff.search", api_call_args)
        diff_list = response.get("data", [])

        diff_dict = {}
        for d in diff_list:
            diff_dict[d["phid"]] = d

        return diff_dict

    def get_successor_phids(
        self, phid: str, include_abandoned: bool = False
    ) -> List[str]:
        return self.get_related_phids(
            phid, relation="child", include_abandoned=include_abandoned
        )

    def get_ancestor_phids(
        self, phid: str, include_abandoned: bool = False
    ) -> List[str]:
        return self.get_related_phids(
            phid, relation="parent", include_abandoned=include_abandoned
        )

    def get_related_phids(
        self, base_phid: str, relation: str = "parent", include_abandoned: bool = False
    ) -> List[str]:
        """Returns the list of PHIDs with direct dependency"""
        result = []

        def _get_related(phid):
            api_call_args = {"sourcePHIDs": [phid], "types": ["revision.%s" % relation]}
            edge = self.call("edge.search", api_call_args)
            if edge.get("data"):
                if len(edge["data"]) > 1:
                    raise NonLinearException()

                result.append(edge["data"][0]["destinationPHID"])
                _get_related(result[-1])

        _get_related(base_phid)

        if not result or include_abandoned:
            return result

        revisions = self.get_revisions(phids=result)
        return [
            r["phid"]
            for r in revisions
            if r["fields"]["status"]["value"] != "abandoned"
        ]

    def get_stack(self, rev_ids: List[int]) -> Dict[str, Any]:
        """Returns a dict of PHIDs."""
        phids = set()
        if not rev_ids:
            return {}
        revisions = self.get_revisions(ids=rev_ids)
        new_phids = set([rev["phid"] for rev in revisions])
        stack = {}

        while new_phids:
            phids.update(new_phids)

            edges = self.call(
                "edge.search",
                dict(
                    sourcePHIDs=list(new_phids),
                    types=["revision.parent", "revision.child"],
                    limit=10000,
                ),
            )["data"]

            new_phids = set()
            for edge in edges:
                new_phids.add(edge["sourcePHID"])
                new_phids.add(edge["destinationPHID"])

                if edge["edgeType"] == "revision.child":
                    if edge["sourcePHID"] in stack:
                        source_id = next(
                            r["id"]
                            for r in revisions
                            if r["phid"] == edge["sourcePHID"]
                        )
                        raise Error("Revision D%s has multiple children." % source_id)

                    stack[edge["sourcePHID"]] = edge["destinationPHID"]

            new_phids = new_phids - phids

        for child in list(stack.values()):
            # set the last child (not a parent)
            stack.setdefault(child)

        return stack

    def get_users(self, usernames: List[str]) -> List[dict]:
        """Get users using the user.query API.

        Caches the result in the process.
        Returns a list of existing Phabricator users data.
        """
        to_collect = []
        users = []
        for user in usernames:
            u = user.rstrip("!")
            key = "user-%s" % u
            if key in cache:
                users.append(cache.get(key))
            else:
                to_collect.append(u)

        if not to_collect:
            return users

        api_call_args = {"usernames": to_collect}
        # We're using the deprecated user.query API as the user.search does not
        # provide the user availability information.
        # See https://phabricator.services.mozilla.com/conduit/method/user.query/
        response = self.call("user.query", api_call_args)
        for user in response:
            users.append(user)
            key = "user-%s" % user["userName"]
            cache.set(key, user)
            cache.set(user["phid"], key)

        return users

    def get_groups(self, slugs: List[str]) -> List[dict]:
        to_collect = []
        groups = []
        for slug in slugs:
            s = slug.rstrip("!")
            key = "group-%s" % s
            if key in cache:
                groups.append(cache.get(key))
            else:
                to_collect.append(s)

        if not to_collect:
            return groups

        # See https://phabricator.services.mozilla.com/conduit/method/project.search/
        api_call_args = {"queryKey": "active", "constraints": {"slugs": to_collect}}
        response = self.call("project.search", api_call_args)
        for data in response.get("data"):
            group = dict(name=data["fields"]["slug"], phid=data["phid"])
            groups.append(group)
            key = "group-%s" % group["name"]
            cache.set(key, group)

        # projects might be received by an alias.
        maps = response["maps"]["slugMap"]
        for alias in maps.keys():
            name = normalise_reviewer(alias)
            group = dict(name=name, phid=maps[alias]["projectPHID"])
            key = "group-%s" % alias
            if key not in cache:
                groups.append(group)
                cache.set(key, group)

        return groups

    def create_revision(
        self,
        commit: dict,
        summary: str,
        diff_phid: str,
        check_in_needed: bool = False,
    ) -> dict:
        """Create a new revision in Phabricator."""
        transactions = [
            dict(type="title", value=revision_title_from_commit(commit)),
            dict(type="summary", value=summary),
        ]
        if commit["has-reviewers"] and not commit["wip"]:
            self.update_revision_reviewers(transactions, commit)

        if commit["bug-id"]:
            transactions.append(dict(type="bugzilla.bug-id", value=commit["bug-id"]))
        return self.edit_revision(
            transactions=transactions,
            diff_phid=diff_phid,
            wip=commit["wip"],
            check_in_needed=check_in_needed,
        )

    def update_revision(
        self,
        commit: dict,
        has_existing_reviewers: bool,
        diff_phid: Optional[str] = None,
        comment: Optional[str] = None,
        check_in_needed: bool = False,
    ) -> dict:
        """Update an existing revision in Phabricator."""
        # Update the title and summary
        transactions = [
            dict(type="title", value=revision_title_from_commit(commit)),
            dict(type="summary", value=strip_differential_revision(commit["body"])),
        ]

        # Add update comment
        if comment:
            transactions.append(dict(type="comment", value=comment))

        # Add reviewers only if revision lacks them
        if commit["has-reviewers"] and not commit["wip"]:
            if not has_existing_reviewers:
                self.update_revision_reviewers(transactions, commit)

        # Update bug id if different
        if commit["bug-id"]:
            revision = conduit.get_revisions(ids=[int(commit["rev-id"])])[0]
            if revision["fields"]["bugzilla.bug-id"] != commit["bug-id"]:
                transactions.append(
                    dict(type="bugzilla.bug-id", value=commit["bug-id"])
                )

        return self.edit_revision(
            transactions=transactions,
            diff_phid=diff_phid,
            rev_id=commit["rev-id"],
            wip=commit["wip"],
            check_in_needed=check_in_needed,
        )

    def edit_revision(
        self,
        transactions: Optional[List[dict]] = None,
        diff_phid: Optional[str] = None,
        rev_id: Optional[str] = None,
        wip: bool = False,
        check_in_needed: bool = False,
        preserve_status: bool = False,
    ) -> dict:
        """Edit (create or update) a revision."""
        trans = transactions or []
        post_trans = []

        # diff_phid is not present for changes in revision settings (like WIP)
        if diff_phid:
            trans.append(dict(type="update", value=diff_phid))

        existing_status = None
        if rev_id:
            try:
                args = dict(ids=[int(rev_id)])
            except ValueError:
                args = dict(phids=[rev_id])
            existing_revision = conduit.get_revisions(**args)[0]
            existing_status = existing_revision["fields"]["status"]["value"]

            if preserve_status and not wip and existing_status != "needs-review":
                # Add a post-update transaction to set the status back to its
                # pre-update value. We don't need to do this for `needs-review`
                # since that is the status we are correcting from.
                post_trans.append(
                    dict(
                        type=REVISION_STATUS_TO_TRANSACTION[existing_status],
                        value=True,
                    )
                )

        # Set revision for changes-planned or needs-review as required.
        # Phabricator will throw an error if we attempt to set a status to the same
        # as the current status.
        if wip:
            # Phabricator will automatically set the revision to needs-review
            # after the call to differential.revision.edit. If we are a creating a new
            # revision, we can add the `plan-changes` transaction to our create call.
            # Existing revisions must make a subsequent API call to set the status to
            # `changes-planned` to match our WIP state.
            plan_changes_transaction = dict(type="plan-changes", value=True)
            if existing_status == "changes-planned":
                post_trans.append(plan_changes_transaction)
            else:
                trans.append(plan_changes_transaction)

        # Let the Phabricator sort out the correct status for new revisions to avoid
        # emails being sent before phabbugs has processed the revision.
        # Updates to existing reviews that have been accepted shouldn't be triggered
        # for re-review.
        elif existing_status and existing_status not in ("needs-review", "accepted"):
            trans.append(dict(type="request-review", value=True))

        # Add the check-in-needed tag if required.
        if check_in_needed:
            logger.info(f"Adding #{CHECK_IN_NEEDED} to revision.")
            check_in_needed_phid = conduit.get_project_phid(CHECK_IN_NEEDED)
            if check_in_needed_phid is not None:
                trans.append(dict(type="projects.add", value=[check_in_needed_phid]))
            else:
                logger.error(
                    f"Could not add #{CHECK_IN_NEEDED} to revision. "
                    f"Project #{CHECK_IN_NEEDED} doesn't exist."
                )

        # Call differential.revision.edit
        api_call_args = dict(transactions=trans)
        if rev_id:
            api_call_args["objectIdentifier"] = rev_id
        revision = self.call("differential.revision.edit", api_call_args)
        if not revision:
            raise ConduitAPIError("Can't edit the revision.")

        # Run post-edit transactions if required.
        if post_trans:
            revision = self.call(
                "differential.revision.edit",
                dict(
                    objectIdentifier=rev_id,
                    transactions=post_trans,
                ),
            )

        return revision

    def get_repository(self, call_sign: str) -> dict:
        """Get the repository info from Phabricator."""
        key = "repo-%s" % call_sign
        if key in cache:
            return dict(cache.get(key))

        api_call_args = dict(constraints=dict(callsigns=[call_sign]), limit=1)
        data = self.call("diffusion.repository.search", api_call_args)
        if not data.get("data"):
            raise NotFoundError("Repository %s not found" % call_sign)

        repo = data["data"][0]
        cache.set(key, repo)
        return repo

    def get_repositories_with_tag(self, tag: str) -> dict:
        """Get repository information for repos associated with the given tag."""
        key = f"repos-tag-{tag}"

        entry = cache.get(key)
        if entry:
            return entry

        api_call_args = {
            "constraints": {
                "projects": [tag],
            }
        }

        data = self.call("diffusion.repository.search", api_call_args).get("data")
        if not data:
            raise NotFoundError(f"No repositories found with tag {tag}")

        cache.set(key, data)
        return data

    def create_diff(self, changes: List[Dict[str, Any]], base_revision: str) -> dict:
        creation_method = ["moz-phab", conduit.repo.vcs]
        if conduit.repo.vcs == "git" and conduit.repo.is_cinnabar_required:
            creation_method.append("cinnabar")

        api_call_args = dict(
            changes=changes,
            sourceMachine=self.repo.phab_url,
            sourceControlSystem=self.repo.phab_vcs,
            sourceControlPath="/",
            sourceControlBaseRevision=base_revision,
            creationMethod="-".join(creation_method),
            lintStatus="none",
            unitStatus="none",
            repositoryPHID=self.repo.phid,
            sourcePath=self.repo.path,
            branch="HEAD" if self.repo.phab_vcs == "git" else "default",
        )
        return self.call("differential.creatediff", api_call_args)

    def set_diff_property(self, diff_id: str, commit: dict, message: str):
        data = {
            commit["node"]: {
                "author": commit["author-name"],
                "authorEmail": commit["author-email"],
                "time": commit["author-date-epoch"],
                "summary": revision_title_from_commit(commit),
                "message": message,
                "commit": conduit.repo.get_public_node(commit["node"]),
                "parents": [conduit.repo.get_public_node(commit["parent"])],
            }
        }
        if "tree-hash" in commit:
            data[commit["node"]]["tree"] = commit["tree-hash"]

        if self.repo.phab_vcs == "hg":
            data[commit["node"]]["rev"] = commit["node"]

        api_call_args = dict(
            diff_id=diff_id, name="local:commits", data=json.dumps(data)
        )
        self.call("differential.setdiffproperty", api_call_args)

    def file_upload(self, path: str, data: bytes) -> Optional[str]:
        if not data:
            return
        name = os.path.basename(path)
        allocation = self.call(
            "file.allocate",
            dict(
                name=name,
                contentLength=len(data),
                contentHash=hashlib.sha256(data).hexdigest(),
            ),
        )
        file_phid = allocation["filePHID"]
        if allocation["upload"]:
            if not file_phid:
                data_base64 = base64.standard_b64encode(data)
                file_phid = self.call(
                    "file.upload", dict(data_base64=data_base64.decode(), name=name)
                )
            else:
                chunks = self.call("file.querychunks", dict(filePHID=file_phid))
                for chunk in chunks:
                    if chunk["complete"]:
                        continue
                    byte_start = int(chunk["byteStart"])
                    byte_end = int(chunk["byteEnd"])
                    data_base64 = base64.standard_b64encode(data[byte_start:byte_end])
                    self.call(
                        "file.uploadchunk",
                        dict(
                            filePHID=file_phid,
                            byteStart=byte_start,
                            data=data_base64.decode(),
                            dataEncoding="base64",
                        ),
                    )

        return str(file_phid)

    def whoami(self, *, api_token: Optional[str] = None) -> dict:
        if "whoami" in cache:
            return dict(cache.get("whoami"))

        who = self.call("user.whoami", {}, api_token=api_token)
        cache.set("whoami", who)
        return who

    def update_revision_reviewers(
        self, transactions: List[Dict[str, Any]], commit: dict
    ):
        # Appends differential.revision.edit transaction(s) to `transactions` to
        # set the reviewers.

        all_reviewing = commit["reviewers"]["request"] + commit["reviewers"]["granted"]

        # Find reviewers PHIDs
        all_reviewers = [r for r in all_reviewing if not r.startswith("#")]
        # preload all reviewers
        self.get_users(all_reviewers)
        reviewers = [r for r in all_reviewers if not r.endswith("!")]
        blocking_reviewers = [r.rstrip("!") for r in all_reviewers if r.endswith("!")]
        reviewers_phid = [user["phid"] for user in self.get_users(reviewers)]
        blocking_phid = [
            "blocking(%s)" % user["phid"] for user in self.get_users(blocking_reviewers)
        ]

        # Find groups PHIDs
        all_groups = [g for g in all_reviewing if g.startswith("#")]
        groups = [g for g in all_groups if not g.endswith("!")]
        blocking_groups = [g.rstrip("!") for g in all_groups if g.endswith("!")]
        # preload all groups
        self.get_groups(all_groups)
        groups_phid = [group["phid"] for group in self.get_groups(groups)]
        bl_groups_phid = [
            "blocking(%s)" % group["phid"] for group in self.get_groups(blocking_groups)
        ]

        all_reviewing_phid = (
            reviewers_phid + blocking_phid + groups_phid + bl_groups_phid
        )
        transactions.extend([dict(type="reviewers.set", value=all_reviewing_phid)])

    def check_for_invalid_reviewers(self, reviewers: dict) -> List[Dict[str, Any]]:
        """Return a list of invalid reviewer names.

        Args:
            reviewers: A commit reviewers dict of granted and requested reviewers.
        """

        # Combine the lists of requested reviewers and granted reviewers.
        all_reviewers = []
        found_names = []
        for sublist in list(reviewers.values()):
            all_reviewers.extend(
                [
                    normalise_reviewer(r, strip_group=False)
                    for r in sublist
                    if not r.startswith("#")
                ]
            )

        users = []
        if all_reviewers:
            users = self.get_users(all_reviewers)
            found_names = [
                normalise_reviewer(data["userName"], strip_group=False)
                for data in users
            ]

        # Group reviewers are represented by a "#" prefix
        all_groups = []
        found_groups = []
        for sublist in list(reviewers.values()):
            all_groups.extend(
                [
                    normalise_reviewer(r, strip_group=False)
                    for r in sublist
                    if r.startswith("#")
                ]
            )

        if all_groups:
            groups = self.get_groups(all_groups)
            found_groups = [
                "#%s" % normalise_reviewer(group["name"]) for group in groups
            ]

        all_reviewers.extend(all_groups)
        found_names.extend(found_groups)
        invalid = list(set(all_reviewers) - set(found_names))

        # Find users availability:
        unavailable = [
            dict(
                name=r["userName"],
                until=datetime.datetime.fromtimestamp(r["currentStatusUntil"]).strftime(
                    "%Y-%m-%d %H:%M"
                ),
            )
            for r in users
            if r.get("currentStatus") == "away"
        ]

        # Find disabled users:
        disabled = [
            dict(name=r["userName"], disabled=True)
            for r in users
            if "disabled" in r.get("roles", [])
        ]
        return disabled + unavailable + [dict(name=r) for r in invalid]


conduit = ConduitAPI()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from pathlib import Path

# Find all command modules

__all__ = sorted(
    [
        f.stem
        for f in Path(__file__).parent.glob("*.py")
        if f.is_file() and f.stem != "__init__"
    ]
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import json
import os
import urllib.parse

from typing import (
    List,
    Optional,
)

from mozphab import environment

from .conduit import conduit, normalise_reviewer
from .exceptions import Error
from .helpers import (
    get_arcrc_path,
    has_arc_rejections,
    read_json_field,
)
from .logger import logger
from .spinner import wait_message


class Repository(object):
    def __init__(self, path, dot_path, phab_url=None):
        self._phid = None
        self._phab_repo = None
        self._phab_vcs = None
        self.vcs = None
        self.path = path  # base repository directory
        self.dot_path = dot_path  # .hg/.git directory
        self._arcconfig_files = [
            os.path.join(self.dot_path, ".arcconfig"),
            os.path.join(self.path, ".arcconfig"),
        ]
        self.args = None
        self.phab_url = (phab_url or self._phab_url()).rstrip("/")
        self.api_url = self._api_url()
        self.call_sign = self._get_setting("repository.callsign")
        self.bmo_url = self._get_setting("bmo_url")

        if self.bmo_url and not (
            urllib.parse.urlparse(self.bmo_url).scheme == "https"
            or environment.HTTP_ALLOWED
        ):
            raise Error("Only https connections are allowed.")

    def is_worktree_clean(self):
        """Check if the working tree is clean."""

    def before_submit(self):
        """Executed before the submit commit."""

    def after_submit(self):
        """Executed after the submit commit."""

    def _get_setting(self, key):
        """Read settings from .arcconfig"""
        value = read_json_field(self._arcconfig_files, [key])
        return value

    def _phab_url(self):
        """Determine the phab/conduit URL."""

        # In order of priority as per arc
        # FIXME: This should also check {.hg|.git}/arc/config, which is where
        # `arc set-config --local` writes to.  See bug 1497786.
        defaults_files = [get_arcrc_path()]
        if environment.IS_WINDOWS:
            defaults_files.append(
                os.path.join(
                    os.getenv("ProgramData", ""), "Phabricator", "Arcanist", "config"
                )
            )
        else:
            defaults_files.append("/etc/arcconfig")

        phab_url = (
            self._get_setting("phabricator.uri")
            or self._get_setting("conduit_uri")
            or read_json_field(defaults_files, ["config", "default"])
        )

        if not phab_url:
            raise Error("Failed to determine Phabricator URL (missing .arcconfig?)")
        return phab_url

    def cleanup(self):
        """Perform any repo-related cleanup tasks.

        May be called multiple times.
        If an exception is raised this is NOT called (to avoid dataloss)."""

    def finalize(self, commits):
        """Update the history after node changed."""

    def set_args(self, args):
        if (
            hasattr(args, "single")
            and args.single
            and args.end_rev != environment.DEFAULT_END_REV
        ):
            raise Error("Option --single can be used with only one identifier.")

        self.args = args

    def untracked(self):
        """Return a list of untracked files."""

    def commit_stack(self, **kwargs):
        """Return list of commits.

        List of dicts with the following keys:
            name          human readable identifier of commit (eg. short sha)
            node          SHA1 in stack
            orig-node     an original SHA1 of the commit
            title         first line of commit description (unaltered)
            body          commit description, excluding first line
            title-preview title with bug-id and reviewer modifications
            bug-id        bmo bug-id
            bug-id-orig   original bug-id from commit desc
            reviewers     list of reviewers
            rev-id        phabricator revision id
            parent        SHA1 of the parent commit
            author-date   string representation of the commit creation time
            author-date-epoch
            author-name
            author-email
        """

    def refresh_commit_stack(self, commits):
        """Update the stack following an altering change (eg rebase)."""

    def is_node(self, node):
        """Check if node exists.

        Returns a Boolean.
        """

    def check_node(self, node):
        """Check if node exists.

        Returns a node if found.

        Raises NotFoundError if node not found in the repository.
        """

    def checkout(self, node):
        """Checkout/Update to specified node."""

    def commit(self, body):
        """Commit the changes in the working directory."""

    def amend_commit(self, commit, commits):
        """Amend commit description from `title` and `desc` fields"""

    def is_descendant(self, node: str) -> bool:
        """Return `True` if the repository revset is descendant from `node`."""

    def map_callsign_to_unified_head(self, callsign: str) -> Optional[str]:
        """Return the expected VCS identifier for the given callsign.

        Returns a VCS identifier that corresponds to the given Phabricator repository
        callsign. Confirms the identified head exists in the repository.
        """

    def uplift_commits(self, dest: str, commits: List[dict]) -> List[dict]:
        """Uplift the repo's revset onto `dest` and returns the refreshed `commits`."""

    def rebase_commit(self, source_commit, dest_commit):
        """Rebase source onto destination."""

    def before_patch(self, node, name):
        """Prepare repository to apply the patches."""

    def apply_patch(self, diff, body, author, author_date):
        """Apply the patch and commit the changes."""

    def format_patch(self, diff, body, author, author_date):
        """Format a patch appropriate for importing."""

    def check_commits_for_submit(self, commits, *, require_bug=True):
        """Validate the list of commits (from commit_stack) are ok to submit"""
        errors = []
        warnings = []

        # Extract a set of reviewers and verify first; they will be displayed
        # with other commit errors.
        all_reviewers = {}
        reviewer_commit_map = {}
        commit_invalid_reviewers = {}
        rev_ids_to_names = dict()
        for commit in commits:
            commit_invalid_reviewers[commit["node"]] = []

            if not commit["rev-id"]:
                continue
            names = rev_ids_to_names.setdefault(commit["rev-id"], [])
            names.append(commit["name"])

        for rev_id, names in rev_ids_to_names.items():
            if len(names) < 2:
                continue

            error_msg = (
                "Phabricator revisions should be unique, but the following "
                "commits refer to the same one (D{}):\n".format(rev_id)
            )
            for name in names:
                error_msg += "* %s\n" % name
            errors.append(error_msg)

        # Flatten and deduplicate reviewer list, keeping track of the
        # associated commit
        for commit in commits:
            # We can ignore reviewers on WIP commits, as they won't be passed to Phab
            if commit["wip"]:
                continue

            for group in list(commit["reviewers"].keys()):
                for reviewer in commit["reviewers"][group]:
                    all_reviewers.setdefault(group, set())
                    all_reviewers[group].add(reviewer)

                    reviewer = normalise_reviewer(reviewer)
                    reviewer_commit_map.setdefault(reviewer, [])
                    reviewer_commit_map[reviewer].append(commit["node"])

        # Verify all reviewers in a single call
        for invalid_reviewer in conduit.check_for_invalid_reviewers(all_reviewers):
            for node in reviewer_commit_map[
                normalise_reviewer(invalid_reviewer["name"])
            ]:
                commit_invalid_reviewers[node].append(invalid_reviewer)

        unavailable_reviewers_warning = False
        for commit in commits:
            commit_errors = []
            commit_warnings = []

            if require_bug and not commit["bug-id"]:
                commit_errors.append("missing bug-id")
            if has_arc_rejections(commit["body"]):
                commit_errors.append("contains arc fields")

            if commit["rev-id"]:
                revisions = conduit.get_revisions(ids=[int(commit["rev-id"])])
                if len(revisions) == 0:
                    commit_errors.append(
                        "Phabricator did not return a query result for revision D%s"
                        " (it might be inaccessible or not exist at all)"
                        % commit["rev-id"]
                    )

            # commit_issues identified below this are commit_errors unless
            # self.args.force is True, which makes them commit_warnings
            commit_issues = (
                commit_warnings if self.args and self.args.force else commit_errors
            )

            for reviewer in commit_invalid_reviewers[commit["node"]]:
                if "disabled" in reviewer:
                    commit_errors.append("User %s is disabled" % reviewer["name"])
                elif "until" in reviewer:
                    unavailable_reviewers_warning = True
                    msg = "%s is not available until %s" % (
                        reviewer["name"],
                        reviewer["until"],
                    )
                    commit_issues.append(msg)
                else:
                    commit_errors.append(
                        "%s is not a valid reviewer's name" % reviewer["name"]
                    )

            if commit_errors:
                errors.append(
                    "%s %s\n- %s"
                    % (commit["name"], commit["title"], "\n- ".join(commit_errors))
                )

            if commit_warnings:
                warnings.append(
                    "%s %s\n- %s"
                    % (commit["name"], commit["title"], "\n- ".join(commit_warnings))
                )

        if errors:
            raise Error("\n\n".join(errors))

        if warnings:
            logger.warning("\n\n".join(warnings))

        if unavailable_reviewers_warning:
            logger.warning("Notice: reviewer availability overridden.")

    def _api_url(self):
        """Return a base URL for conduit API call"""
        url = urllib.parse.urljoin(self.phab_url, "api/")

        if not (
            urllib.parse.urlparse(url).scheme == "https" or environment.HTTP_ALLOWED
        ):
            raise Error("Only https connections are allowed.")

        return url

    @property
    def phab_repo(self):
        """Representation of the Repository in Phabricator API."""
        if not self._phab_repo:
            with wait_message("Reading repository data"):
                self._phab_repo = conduit.get_repository(self.call_sign)

        return self._phab_repo

    @property
    def phid(self):
        """PHID of the repository.

        This value does not change over time.
        It is stored in a file to avoid calling the API on every run.
        """
        if not self._phid:
            path = os.path.join(self.dot_path, ".moz-phab_phid")

            if os.path.isfile(path):
                with open(path) as f:
                    try:
                        repo_phids = json.load(f)
                    except json.decoder.JSONDecodeError:
                        # File is probably using the old format.
                        repo_phids = {}
                    repo_phid = repo_phids.get(self.call_sign, None)
            else:
                repo_phids = {}
                repo_phid = None

            if not repo_phid:
                repo_phid = self.phab_repo["phid"]
                repo_phids[self.call_sign] = repo_phid
                with open(path, "w") as f:
                    json.dump(repo_phids, f)
            self._phid = repo_phid

        return self._phid

    def check_vcs(self):
        """`Git.check_vcs` raises if cinnabar required and not installed."""
        if self.args.force_vcs:
            return True

        if self.vcs != self.phab_vcs:
            # This error is captured in Git and not raised if Cinnabar installed.
            raise Error(
                "Local VCS ({local}) is different from the one defined in the "
                "repository ({remote}).".format(local=self.vcs, remote=self.phab_vcs)
            )

        return True

    @property
    def phab_vcs(self):
        """Version Control System short name stored in Phabricator.

        This value does not change over time.
        It is stored in a file to avoid calling the API on every run.
        """
        if not self._phab_vcs:
            # check file
            path = os.path.join(self.dot_path, ".moz-phab_vcs")
            if os.path.isfile(path):
                with open(path) as f:
                    self._phab_vcs = f.readline()
            else:
                self._phab_vcs = self.phab_repo["fields"]["vcs"]
                with open(path, "w") as f:
                    f.write(self._phab_vcs)

        return self._phab_vcs

    def get_public_node(self, node):
        """Hashtag in a remote VCS."""
        return node

    def validate_email(self):
        """Validate a user's configured email address."""



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import copy
import pytest
from packaging.version import Version

from unittest import mock

from .conftest import create_temp_fn, assert_attributes

from mozphab import environment, exceptions, mozphab, diff
from mozphab.diff import Diff
from mozphab.mercurial import Mercurial


@mock.patch("mozphab.mercurial.Mercurial.hg_out")
def test_get_successor(m_hg_hg_out, hg):
    m_hg_hg_out.return_value = []
    assert (None, None) == hg._get_successor("x")

    m_hg_hg_out.return_value = ["1 abcde"]
    assert ["1", "abcde"] == hg._get_successor("x")

    m_hg_hg_out.return_value = ["a", "b"]
    with pytest.raises(exceptions.Error):
        hg._get_successor("x")


@mock.patch("mozphab.mercurial.Mercurial._get_successor")
@mock.patch("mozphab.mercurial.Mercurial.rebase_commit")
@mock.patch("mozphab.mercurial.Mercurial._get_parent")
def test_finalize(m_get_parent, m_hg_rebase, m_hg_get_successor, hg):
    commits = [
        {"rev": "1", "node": "aaa", "orig-node": "aaa"},
        {"rev": "2", "node": "bbb", "orig-node": "bbb"},
        {"rev": "3", "node": "ccc", "orig-node": "ccc"},
    ]

    m_get_parent.return_value = "different:than_others"
    m_hg_get_successor.return_value = (None, None)
    hg.finalize(copy.deepcopy(commits))
    assert m_hg_rebase.call_count == 2
    assert m_hg_rebase.call_args_list == [
        mock.call(
            {"rev": "2", "node": "bbb", "orig-node": "bbb"},
            {"rev": "1", "node": "aaa", "orig-node": "aaa"},
        ),
        mock.call(
            {"rev": "3", "node": "ccc", "orig-node": "ccc"},
            {"rev": "2", "node": "bbb", "orig-node": "bbb"},
        ),
    ]

    m_get_parent.side_effect = ("first", "aaa", "last")
    m_hg_rebase.reset_mock()
    hg.finalize(commits)
    m_hg_rebase.assert_called_once_with(
        {"rev": "3", "node": "ccc", "orig-node": "ccc"},
        {"rev": "2", "node": "bbb", "orig-node": "bbb"},
    )

    m_hg_get_successor.reset_mock()
    m_get_parent.side_effect = None
    m_get_parent.return_value = "different:than_others"
    m_hg_get_successor.side_effect = [(None, None), ("4", "ddd")]
    _commits = commits[:]
    hg.finalize(_commits)
    assert m_hg_get_successor.call_count == 2
    assert m_hg_get_successor.call_args_list == [mock.call("bbb"), mock.call("ccc")]
    assert _commits == [
        {"rev": "1", "node": "aaa", "orig-node": "aaa"},
        {"rev": "2", "node": "bbb", "orig-node": "bbb"},
        {"rev": "3", "node": "ddd", "orig-node": "ccc", "name": "4:ddd"},
    ]

    m_hg_rebase.reset_mock()
    m_hg_get_successor.side_effect = None
    m_hg_get_successor.return_value = (None, None)
    _commits = commits[:]
    _commits[0]["node"] = "AAA"  # node has been amended
    hg.finalize(_commits)
    assert m_hg_rebase.call_count == 2


@mock.patch("mozphab.mercurial.Mercurial.rebase_commit")
def test_finalize_no_evolve(m_hg_rebase, hg):
    hg.use_evolve = False
    hg.finalize([dict(rev="1", node="aaa"), dict(rev="2", node="bbb")])
    assert m_hg_rebase.not_called()


@mock.patch("mozphab.mercurial.parse_config")
@mock.patch("mozphab.mercurial.Mercurial.hg_out")
@mock.patch("mozphab.mercurial.Mercurial.hg_log")
@mock.patch("mozphab.mercurial.hglib.open")
def test_set_args(m_hglib_open, m_hg_hg_log, m_hg_hg_out, m_parse_config, hg):
    class Args:
        def __init__(self, start="(auto)", end=".", safe_mode=False, single=False):
            self.start_rev = start
            self.end_rev = end
            self.safe_mode = safe_mode
            self.single = single

    m_config = mozphab.config
    with pytest.raises(exceptions.Error):
        hg.set_args(Args())

    # baseline config
    hg.mercurial_version = Version("4.5")
    m_config.safe_mode = False
    m_parse_config.return_value = {
        "ui.username": "username",
        "extensions.evolve": "",
        "extensions.topic": "",
    }

    # evolve & topic & shelve
    hg._hg = []
    hg.set_args(Args())

    assert "extensions.rebase" in hg._config_options
    assert hg._config_options["rebase.experimental.inmemory"] == "true"
    assert hg._extra_options["--pager"] == "never"
    assert hg.use_evolve
    assert hg.use_topic
    assert not hg.has_shelve

    # inmemory rebase requires hg 4.5+
    hg.mercurial_version = Version("4.0")
    hg._hg = []
    hg.set_args(Args())
    assert "extensions.rebase" in hg._config_options
    assert "rebase.experimental.inmemory" not in hg._config_options
    assert hg._extra_options["--pager"] == "never"
    hg.mercurial_version = Version("4.5")

    # safe_mode
    hg._hg = []
    hg.set_args(Args(safe_mode=True))
    options = hg._get_config_options()
    assert options == [
        ("extensions.rebase", ""),
        ("ui.username", "username"),
        ("extensions.evolve", ""),
        ("extensions.topic", ""),
    ]

    m_config.safe_mode = True
    hg._hg = []
    hg.set_args(Args())
    options = hg._get_config_options()
    assert options == [
        ("extensions.rebase", ""),
        ("ui.username", "username"),
        ("extensions.evolve", ""),
        ("extensions.topic", ""),
    ]
    m_config.safe_mode = False

    # no evolve
    m_parse_config.return_value = {"ui.username": "username", "extensions.shelve": ""}
    hg._hg = []
    hg.set_args(Args())
    options = hg._get_config_options()
    assert options == [
        ("extensions.rebase", ""),
        ("rebase.experimental.inmemory", "true"),
        ("experimental.evolution.createmarkers", "true"),
        ("extensions.strip", ""),
    ]
    assert not hg.use_evolve
    assert hg.has_shelve

    m_hg_hg_log.side_effect = [("123456789012",), ("098765432109",)]
    hg._hg = []
    hg.set_args(Args())
    assert "123456789012::098765432109" == hg.revset

    m_hg_hg_log.side_effect = IndexError
    with pytest.raises(exceptions.Error):
        hg.set_args(Args())

    m_hg_hg_log.reset_mock()
    m_hg_hg_log.side_effect = [("123456789012",), ("123456789012",)]
    hg.set_args(Args(single=True))
    assert "123456789012" == hg.revset
    assert m_hg_hg_log.call_args_list == [mock.call(".")]

    m_hg_hg_log.reset_mock()
    m_hg_hg_log.side_effect = [("123456789012",), ("123456789012",)]
    hg.set_args(Args(start="start", single=True))
    assert "123456789012" == hg.revset
    assert m_hg_hg_log.call_args_list == [mock.call("start")]


@mock.patch("mozphab.mercurial.Mercurial._status")
def test_clean_worktree(m_status, hg):
    m_status.return_value = {"T": None, "U": None}
    assert hg.is_worktree_clean()

    m_status.return_value = {"T": True, "U": None}
    assert not hg.is_worktree_clean()

    m_status.return_value = {"T": None, "U": True}
    assert hg.is_worktree_clean()

    m_status.return_value = {"T": True, "U": True}
    assert not hg.is_worktree_clean()


@mock.patch("mozphab.mercurial.Mercurial.hg")
def test_commit(m_hg, hg):
    hg.commit("some body")
    m_hg.called_once()


@mock.patch("mozphab.mercurial.Mercurial.checkout")
@mock.patch("mozphab.mercurial.Mercurial.hg_out")
@mock.patch("mozphab.mercurial.Mercurial.hg")
@mock.patch("mozphab.mercurial.config")
def test_before_patch(m_config, m_hg, m_hg_out, m_checkout, hg):
    class Args:
        def __init__(
            self,
            rev_id="D123",
            nocommit=False,
            raw=False,
            applyto="base",
            no_bookmark=False,
            no_topic=False,
        ):
            self.rev_id = rev_id
            self.nocommit = nocommit
            self.raw = raw
            self.applyto = applyto
            self.no_bookmark = no_bookmark
            self.no_topic = no_topic

    m_config.create_bookmark = True
    m_config.create_topic = False
    m_hg_out.side_effect = ["bookmark"]
    hg.args = Args()
    hg.before_patch("sha1", "bookmark")
    m_checkout.assert_called_with("sha1")
    m_hg_out.assert_called()
    m_hg.assert_called_with(["bookmark", "bookmark_1"])
    m_checkout.assert_called_once_with("sha1")

    m_checkout.reset_mock()
    hg.args = Args(nocommit=True)
    m_hg.reset_mock()
    hg.before_patch("sha1", None)
    m_hg.assert_not_called()
    m_checkout.assert_called_once_with("sha1")

    hg.args = Args(applyto="here")
    m_checkout.reset_mock()
    m_hg_out.reset_mock()
    m_hg_out.side_effect = None
    m_hg_out.return_value = "some book_marks"
    hg.before_patch(None, "bookmark")
    m_hg_out.assert_called_once()
    m_checkout.assert_not_called()

    hg.args = Args(applyto="here")
    m_checkout.reset_mock()
    m_hg_out.reset_mock()
    m_hg_out.side_effect = None
    m_hg_out.return_value = "some book_marks"
    hg.before_patch(None, "bookmark")
    m_hg_out.assert_called_once()
    m_checkout.assert_not_called()

    m_hg_out.reset_mock()
    hg.args = Args(no_bookmark=True)
    hg.before_patch(None, "bookmark")
    m_hg_out.assert_not_called()

    m_config.create_bookmark = False
    m_hg_out.reset_mock()
    hg.args = Args()
    hg.before_patch(None, "bookmark")
    m_hg_out.assert_not_called()

    # Create a topic, but requested topic collides with an existing
    # topic and so must be renamed.
    m_config.create_topic = True
    hg.use_topic = True
    m_hg_out.reset_mock()
    m_hg_out.side_effect = ["topic_1"]
    hg.args = Args()
    hg.before_patch("sha1", "topic")
    m_checkout.assert_called_with("sha1")
    m_hg_out.assert_called()
    m_hg.assert_called_with(["topic", "topic_2"])
    m_checkout.assert_called_once_with("sha1")

    # No conflict => use requested name.
    m_hg_out.side_effect = ["topic"]
    hg.before_patch("sha1", "other")
    m_hg.assert_called_with(["topic", "other"])

    # Create both a bookmark and a topic. Not a useful thing to do,
    # but it should work. And bookmark names and topic names should not
    # collide when looking for pre-existing names.

    m_config.create_bookmark = True
    m_hg.reset_mock()  # Clear the calls.
    m_hg_out.side_effect = ["bookmark", "topic"]
    hg.before_patch("sha1", "bookmark")
    m_hg.assert_has_calls(
        [
            mock.call(["bookmark", "bookmark_1"]),
            mock.call(["topic", "bookmark"]),
        ]
    )


@mock.patch("mozphab.mercurial.temporary_binary_file")
@mock.patch("mozphab.mercurial.Mercurial.hg")
def test_apply_patch(m_hg, m_temp_bin_fn, hg):
    m_temp_bin_fn.return_value = create_temp_fn("diff_fn")
    hg.apply_patch("diff", "body", "user", 1)
    m_hg.assert_called_once_with(["import", "diff_fn", "--quiet"])
    m_temp_bin_fn.assert_called_once_with(
        b"# HG changeset patch\n# User user\n# Date 1 0\nbody\ndiff"
    )


@mock.patch("mozphab.mercurial.Mercurial.hg_out")
def test_is_node(m_hg_out, hg):
    assert hg.is_node("aabbcc")
    m_hg_out.assert_called_once_with(["identify", "-q", "-r", "aabbcc"])

    m_hg_out.side_effect = mock.Mock(side_effect=exceptions.CommandError)
    assert not hg.is_node("aaa")


@mock.patch("mozphab.mercurial.Mercurial.hg_out")
def test_is_descendant(m_hg_out, hg):
    m_hg_out.return_value = ""
    assert (
        hg.is_descendant("aabbcc") is False
    ), "Empty log result should indicated commit is not a descendant."

    m_hg_out.return_value = "aabbcc"
    assert (
        hg.is_descendant("aabbcc") is True
    ), "Non-empty log result should indicated commit is a descendant."


@mock.patch("mozphab.mercurial.Mercurial.is_node")
def test_check_node(m_is_node, hg):
    node = "aabbcc"
    m_is_node.return_value = True
    assert node == hg.check_node(node)

    m_is_node.return_value = False
    with pytest.raises(exceptions.NotFoundError) as e:
        hg.check_node(node)

    assert "" == str(e.value)


@mock.patch("mozphab.mercurial.Mercurial.hg_out")
def test_hg_cat(m_hg, hg):
    cat = m_hg.return_value = b"some text"
    hg.hg_cat("fn", "node")
    m_hg.assert_called_once_with(
        ["cat", "-r", "node", "fn"], expect_binary=True, split=False
    )
    assert cat == b"some text"


@mock.patch("mozphab.mercurial.Mercurial.hg_out")
def test_file_size(m_hg, hg):
    m_hg.return_value = "123\n"
    res = hg._file_size("fn", "rev")
    m_hg.assert_called_once_with(
        ["files", "-v", "-r", "rev", mock.ANY, "-T", "{size}"], split=False
    )
    assert res == 123


@mock.patch("mozphab.mercurial.Mercurial._file_size")
@mock.patch("mozphab.mercurial.Mercurial.hg_cat")
def test_file_meta(m_cat, m_file_size, hg):
    size = environment.MAX_TEXT_SIZE - 1
    m_file_size.return_value = size
    m_cat.return_value = b"spam\nham"
    meta = hg._get_file_meta("fn", "rev")
    assert meta == dict(
        binary=False,
        mime="TEXT",
        bin_body=b"spam\nham",
        body="spam\nham",
        file_size=size,
    )


@mock.patch("mozphab.mercurial.mimetypes")
@mock.patch("mozphab.mercurial.Mercurial._file_size")
@mock.patch("mozphab.mercurial.Mercurial.hg_cat")
def test_file_meta_binary(m_cat, m_file_size, m_mime, hg):
    m_mime.guess_type.return_value = ["MIMETYPE"]
    m_cat.return_value = b"spam\nham"
    size = environment.MAX_TEXT_SIZE + 1
    m_file_size.return_value = size
    meta = hg._get_file_meta("fn", "rev")
    assert meta == dict(
        binary=True,
        mime="MIMETYPE",
        bin_body=b"spam\nham",
        body=b"spam\nham",
        file_size=size,
    )

    hg._get_file_meta.cache_clear()
    hg.hg_cat.cache_clear()
    hg._file_size.cache_clear()

    size = environment.MAX_TEXT_SIZE - 1
    m_file_size.return_value = size
    m_cat.return_value = b"\0spam\nham"
    meta = hg._get_file_meta("fn", "rev")
    assert meta == dict(
        binary=True,
        mime="MIMETYPE",
        bin_body=b"\0spam\nham",
        body=b"\0spam\nham",
        file_size=size,
    )


@mock.patch("mozphab.mercurial.Mercurial._get_file_meta")
@mock.patch("mozphab.diff.Diff.Change.set_as_binary")
def test_change_add(m_set_as_binary, m_get_file_meta, hg):
    change = Diff.Change("x")
    m_get_file_meta.return_value = dict(
        binary=False,
        bin_body=b"abc\n",
        body="abc\n",
        file_size=123,
    )
    hg._change_add(change, "fn", None, "parent", "node")
    m_set_as_binary.assert_not_called()
    assert len(change.hunks) == 1
    assert_attributes(
        change.hunks[0],
        dict(
            corpus="+abc\n",
            old_off=0,
            new_off=1,
            old_len=0,
            new_len=1,
        ),
    )

    # binary
    change = Diff.Change("x")
    m_set_as_binary.reset_mock()
    m_get_file_meta.return_value = dict(
        binary=True,
        bin_body=b"abc\n",
        body="abc\n",
        file_size=123,
        mime="MIME",
    )
    hg._change_add(change, "fn", None, "parent", "node")
    assert not change.hunks
    m_set_as_binary.assert_called_once_with(
        a_body="",
        a_mime="",
        b_body=b"abc\n",
        b_mime="MIME",
    )

    # empty
    change = Diff.Change("x")
    m_set_as_binary.reset_mock()
    m_get_file_meta.return_value = dict(
        binary=False,
        bin_body=b"",
        body="",
        file_size=0,
    )
    hg._change_add(change, "fn", None, "parent", "node")
    assert not change.hunks
    m_set_as_binary.assert_not_called()


@mock.patch("mozphab.mercurial.Mercurial._get_file_meta")
@mock.patch("mozphab.diff.Diff.Change.set_as_binary")
def test_change_del(m_set_as_binary, m_get_file_meta, hg):
    change = Diff.Change("x")
    m_get_file_meta.return_value = dict(
        binary=False,
        bin_body=b"abc\n",
        body="abc\n",
        file_size=123,
    )
    hg._change_del(change, "fn", None, "parent", "node")
    assert len(change.hunks) == 1
    assert_attributes(
        change.hunks[0],
        dict(
            corpus="-abc\n",
            old_off=1,
            new_off=0,
            old_len=1,
            new_len=0,
        ),
    )
    m_set_as_binary.assert_not_called()

    # binary
    change = Diff.Change("x")
    m_set_as_binary.reset_mock()
    m_get_file_meta.return_value = dict(
        binary=True,
        bin_body=b"abc\n",
        body="abc\n",
        file_size=123,
        mime="MIME",
    )
    hg._change_del(change, "fn", None, "parent", "node")
    assert not change.hunks
    m_set_as_binary.assert_called_once_with(
        a_body=b"abc\n",
        a_mime="MIME",
        b_body="",
        b_mime="",
    )

    # empty
    change = Diff.Change("x")
    m_set_as_binary.reset_mock()
    m_get_file_meta.return_value = dict(
        binary=False,
        bin_body=b"",
        body="",
        file_size=0,
    )
    hg._change_del(change, "fn", None, "parent", "node")
    assert not change.hunks
    m_set_as_binary.assert_not_called()


@mock.patch("mozphab.mercurial.Mercurial._get_file_meta")
@mock.patch("mozphab.diff.Diff.Change.set_as_binary")
@mock.patch("mozphab.diff.Diff.Change.from_git_diff")
@mock.patch("mozphab.mercurial.Mercurial.hg_out")
def test_change_mod(m_hg_out, m_from_git_diff, m_set_as_binary, m_get_file_meta, hg):
    class Args:
        def __init__(self, lesscontext=False):
            self.lesscontext = lesscontext

    # file contents changed
    change = diff.Diff.Change(None)
    text_side_effect = (
        dict(binary=False, bin_body=b"abc\n", body="abc\n", file_size=4),
        dict(binary=False, bin_body=b"def\n", body="def\n", file_size=4),
    )
    m_get_file_meta.side_effect = text_side_effect
    m_hg_out.return_value = b"""\
diff --git a/fn b/fn
--- a/B
+++ b/B
@@ -1,1 +1,1 @@
-abc
+def
"""
    hg.args = Args()
    hg._change_mod(change, "fn", "old_fn", "parent", "node")
    m_hg_out.assert_called_once_with(
        ["diff"]
        + ["--git"]
        + ["--unified", str(environment.MAX_CONTEXT_SIZE)]
        + ["--rev", "parent"]
        + ["fn"],
        expect_binary=True,
    )
    m_from_git_diff.assert_called_once()

    # --less-content honoured
    m_hg_out.reset_mock()
    m_get_file_meta.side_effect = text_side_effect
    hg.args = Args(lesscontext=True)
    hg._change_mod(change, "fn", "old_fn", "parent", "node")
    m_hg_out.assert_called_once_with(
        ["diff", "--git", "--unified", "100", "--rev", "parent", "fn"],
        expect_binary=True,
    )

    # binary
    m_from_git_diff.reset_mock()
    m_get_file_meta.side_effect = (
        dict(binary=True, bin_body=b"abc\n", body=b"abc\n", file_size=4, mime="MIME"),
        dict(binary=False, bin_body=b"def\n", body="def\n", file_size=4, mime="TEXT"),
    )
    hg._change_mod(change, "fn", "old_fn", "parent", "node")
    m_set_as_binary.assert_called_once_with(
        a_body=b"abc\n", a_mime="MIME", b_body=b"def\n", b_mime="TEXT"
    )


@mock.patch("mozphab.mercurial.Mercurial.hg_out")
def test_get_file_modes(m_hg, hg):
    m_hg.side_effect = (
        ["  file name"],  # status
        [" :file name"],  # files - parent
        [" :file name"],  # files - node
    )
    actual = hg._get_file_modes(dict(node="aaa", parent="bbb"))
    expected = {"file name": dict(old_mode="100644", new_mode="100644")}
    assert actual == expected

    m_hg.side_effect = (
        ["M file name"],  # status
        [" :file name"],  # files - parent
        ["x:file name"],  # files - node
    )
    actual = hg._get_file_modes(dict(node="aaa", parent="bbb"))
    expected = {"file name": dict(old_mode="100644", new_mode="100755")}
    assert actual == expected

    m_hg.side_effect = (
        ["A file name"],  # status
        [],  # files - parent
        ["x:file name"],  # files - node
    )
    actual = hg._get_file_modes(dict(node="aaa", parent="bbb"))
    expected = {"file name": dict(new_mode="100755")}
    assert actual == expected

    m_hg.side_effect = (
        ["R file name"],  # status
        [" :file name"],  # files - parent
        [],  # files - node
    )
    actual = hg._get_file_modes(dict(node="aaa", parent="bbb"))
    expected = {"file name": dict(old_mode="100644")}
    assert actual == expected


def test_check_vcs(hg):
    class Args:
        def __init__(self, force_vcs=False):
            self.force_vcs = force_vcs

    hg.args = Args()
    assert hg.check_vcs()

    hg._phab_vcs = "git"
    with pytest.raises(exceptions.Error):
        hg.check_vcs()

    hg.args = Args(force_vcs=True)
    assert hg.check_vcs()


@mock.patch("mozphab.mercurial.hglib.open")
@mock.patch("mozphab.repository.Repository._phab_url")
@mock.patch("mozphab.repository.os.chdir")
@mock.patch("os.path.isdir")
@mock.patch("mozphab.helpers.which")
def test_repository_cached(m_which, m_is_dir, m_os_chdir, m_phab_url, m_open, *patched):
    class MyRepo:
        version = 4, 7, 3

        def rawcommand(self, *args, **kw):
            return b"ui.username=xxx"

        def close(self):
            pass

    m_is_dir.return_value = True
    m_os_chdir.return_value = True
    m_phab_url.return_value = ""
    m_which.return_value = True
    m_open.return_value = MyRepo()

    hg = Mercurial("x")

    class Args:
        def __init__(self):
            self.start_rev = "(auto)"
            self.end_rev = "."
            self.safe_mode = True
            self.single = False

    hg._repo = None
    hg.set_args(Args())
    current_repo = hg.repository
    assert current_repo is not None

    # makes sure we cache the `hgclient` instance when using the
    # same args
    hg.set_args(Args())
    assert hg.repository is current_repo


@mock.patch("mozphab.mercurial.Mercurial.is_node")
def test_hg_map_callsign_to_unified_head(m_is_node, hg):
    m_is_node.return_value = False
    assert (
        hg.map_callsign_to_unified_head("blah") is None
    ), "Unknown head should have returned `None`."

    m_is_node.return_value = True
    assert (
        hg.map_callsign_to_unified_head("beta") == "beta"
    ), "beta did not correctly map to a branch"


def test_hg_validate_email(hg):
    with pytest.raises(exceptions.Error):
        hg.validate_email()

    hg.username = "Test User <test@mozilla.com>"
    assert hg.validate_email() is None, "validate_email() executes without error"


def test_hg_extract_email_from_username(hg):
    # Empty username
    hg.username = ""
    assert hg.extract_email_from_username() == ""

    # Username without email
    hg.username = "test user"
    assert hg.extract_email_from_username() == "test user"

    # Username with email
    hg.username = "test user <test@email.com>"
    assert hg.extract_email_from_username() == "test@email.com"
